import torch
import torch.nn.functional as F

from .sde import get_score_fn
from .utils import mask_adjs, pow_tensor, mask_x, gen_noise
from .layers import DenseGCNConv, MLP, AttentionLayer

# class GDSSModel(torch.nn.Module):
#     def __init__(self, max_feat_num, max_node_num, nhid, num_layers, num_linears, 
#                     c_init, c_hid, c_final, adim, num_heads=4, conv='GCN'):
#         super(GDSSModel, self).__init__()

#         self.score_network_a = ScoreNetworkA(max_feat_num, max_node_num, nhid, num_layers, num_linears, 
#                                             c_init, c_hid, c_final, adim, num_heads, conv)
#         self.score_network_x = ScoreNetworkX(max_feat_num, num_layers, nhid)

class GDSSModel(torch.nn.Module):
    def __init__(self, input_dim_X, max_node, hidden_size, num_layer, input_dim_adj, hidden_size_adj, attention_dim, num_head=4, conv='GCN'):
        super(GDSSModel, self).__init__()
        num_layer_in_mlp = 3

        self.score_network_a = ScoreNetworkA(
            max_feat_num=input_dim_X,
            max_node_num=max_node,
            nhid=hidden_size,
            num_layers=num_layer,
            num_linears=num_layer_in_mlp,
            c_init=input_dim_adj,
            c_hid=hidden_size_adj,
            c_final=input_dim_adj,
            adim=attention_dim,
            num_heads=num_head,
            conv=conv
        )
        self.score_network_x = ScoreNetworkX(
            max_feat_num=input_dim_X,
            depth=num_layer,
            nhid=hidden_size
        )

    def compute_loss(self, x, adj, flags, loss_fn):
        loss_x, loss_adj = loss_fn(
            self.score_network_x, self.score_network_a, x, adj, flags
        )
        return loss_x, loss_adj
    
    def forward(self, x, adj, flags):
        raise NotImplementedError("Forward pass not implemented. Please use self.score_network_x and self.score_network_a separately.")
    
    def initialize_parameters(self, seed=None):
        if seed is not None:
            torch.manual_seed(seed)

        # Initialize transformer layers:
        def _basic_init(module):
            if isinstance(module, torch.nn.Linear):
                torch.nn.init.xavier_uniform_(module.weight)
                if module.bias is not None:
                    torch.nn.init.constant_(module.bias, 0)

        self.apply(_basic_init)

class ScoreNetworkA(torch.nn.Module):
    def __init__(self, max_feat_num, max_node_num, nhid, num_layers, num_linears, 
                    c_init, c_hid, c_final, adim, num_heads=4, conv='GCN'):
        super(ScoreNetworkA, self).__init__()
        
        self.adim = adim
        self.num_heads = num_heads
        self.conv = conv
        self.num_layers = num_layers
        self.num_linears = num_linears
        self.nfeat = max_feat_num
        self.nhid = nhid
        self.c_init = c_init
        self.c_hid = c_hid
        self.c_final = c_final
        self.max_node_num = max_node_num

        self.layers = torch.nn.ModuleList()
        for i in range(self.num_layers):
            if i==0:
                self.layers.append(AttentionLayer(self.num_linears, self.nfeat, self.nhid, self.nhid, self.c_init, 
                                                    self.c_hid, self.num_heads, self.conv))
            elif i==self.num_layers-1:
                self.layers.append(AttentionLayer(self.num_linears, self.nhid, self.adim, self.nhid, self.c_hid, 
                                                    self.c_final, self.num_heads, self.conv))
            else:
                self.layers.append(AttentionLayer(self.num_linears, self.nhid, self.adim, self.nhid, self.c_hid, 
                                                    self.c_hid, self.num_heads, self.conv))

        self.fdim = self.c_hid*(self.num_layers-1) + self.c_final + self.c_init
        self.final = MLP(num_layers=3, input_dim=self.fdim, hidden_dim=2*self.fdim, output_dim=1, 
                            use_bn=False, activate_func=F.elu)
        self.mask = torch.ones([self.max_node_num, self.max_node_num]) - torch.eye(self.max_node_num)
        self.mask.unsqueeze_(0)  

    def forward(self, x, adj, flags):
        adjc = pow_tensor(adj, self.c_init)

        adj_list = [adjc]
        for i in range(self.num_layers):
            x, adjc = self.layers[i](x, adjc, flags)
            adj_list.append(adjc)
        
        adjs = torch.cat(adj_list, dim=1).permute(0,2,3,1)
        out_shape = adjs.shape[:-1] # B x N x N
        score = self.final(adjs).view(*out_shape)
        
        self.mask = self.mask.to(score.device)
        score = score * self.mask
        score = mask_adjs(score, flags)
        return score

class ScoreNetworkX(torch.nn.Module):
    def __init__(self, max_feat_num, depth, nhid):
        super(ScoreNetworkX, self).__init__()
        self.nfeat = max_feat_num
        self.depth = depth
        self.nhid = nhid
        self.layers = torch.nn.ModuleList()
        for _ in range(self.depth):
            if _ == 0:
                self.layers.append(DenseGCNConv(self.nfeat, self.nhid))
            else:
                self.layers.append(DenseGCNConv(self.nhid, self.nhid))

        self.fdim = self.nfeat + self.depth * self.nhid
        self.final = MLP(num_layers=3, input_dim=self.fdim, hidden_dim=2*self.fdim, output_dim=self.nfeat, 
                            use_bn=False, activate_func=F.elu)

        self.activation = torch.tanh

    def forward(self, x, adj, flags):

        x_list = [x]
        for _ in range(self.depth):
            x = self.layers[_](x, adj)
            x = self.activation(x)
            x_list.append(x)

        xs = torch.cat(x_list, dim=-1) # B x N x (F + num_layers x H)
        out_shape = (adj.shape[0], adj.shape[1], -1)
        x = self.final(xs).view(*out_shape)

        x = mask_x(x, flags)

        return x

def get_sde_loss_fn(
    sde_x,
    sde_adj,
    train=True,
    reduce_mean=False,
    continuous=True,
    likelihood_weighting=False,
    eps=1e-5,
):
    reduce_op = (
        torch.mean
        if reduce_mean
        else lambda *args, **kwargs: 0.5 * torch.sum(*args, **kwargs)
    )

    def loss_fn(model_x, model_adj, x, adj, flags):
        score_fn_x = get_score_fn(sde_x, model_x, train=train, continuous=continuous)
        score_fn_adj = get_score_fn(
            sde_adj, model_adj, train=train, continuous=continuous
        )

        t = torch.rand(adj.shape[0], device=adj.device) * (sde_adj.T - eps) + eps
        z_x = gen_noise(x, flags, sym=False)
        mean_x, std_x = sde_x.marginal_prob(x, t)
        perturbed_x = mean_x + std_x[:, None, None] * z_x
        perturbed_x = mask_x(perturbed_x, flags)

        z_adj = gen_noise(adj, flags, sym=True)
        mean_adj, std_adj = sde_adj.marginal_prob(adj, t)
        perturbed_adj = mean_adj + std_adj[:, None, None] * z_adj
        perturbed_adj = mask_adjs(perturbed_adj, flags)

        score_x = score_fn_x(perturbed_x, perturbed_adj, flags, t)
        score_adj = score_fn_adj(perturbed_x, perturbed_adj, flags, t)

        if not likelihood_weighting:
            losses_x = torch.square(score_x * std_x[:, None, None] + z_x)
            losses_x = reduce_op(losses_x.reshape(losses_x.shape[0], -1), dim=-1)

            losses_adj = torch.square(score_adj * std_adj[:, None, None] + z_adj)
            losses_adj = reduce_op(losses_adj.reshape(losses_adj.shape[0], -1), dim=-1)

        else:
            g2_x = sde_x.sde(torch.zeros_like(x), t)[1] ** 2
            losses_x = torch.square(score_x + z_x / std_x[:, None, None])
            losses_x = reduce_op(losses_x.reshape(losses_x.shape[0], -1), dim=-1) * g2_x

            g2_adj = sde_adj.sde(torch.zeros_like(adj), t)[1] ** 2
            losses_adj = torch.square(score_adj + z_adj / std_adj[:, None, None])
            losses_adj = (
                reduce_op(losses_adj.reshape(losses_adj.shape[0], -1), dim=-1) * g2_adj
            )
        return torch.mean(losses_x), torch.mean(losses_adj)

    return loss_fn



import rdkit
import rdkit.Chem as Chem
import copy

def get_slots(smiles):
    mol = Chem.MolFromSmiles(smiles)
    return [(atom.GetSymbol(), atom.GetFormalCharge(), atom.GetTotalNumHs()) for atom in mol.GetAtoms()]

class Vocab(object):
    benzynes = ['C1=CC=CC=C1', 'C1=CC=NC=C1', 'C1=CC=NN=C1', 'C1=CN=CC=N1', 'C1=CN=CN=C1', 'C1=CN=NC=N1', 'C1=CN=NN=C1', 'C1=NC=NC=N1', 'C1=NN=CN=N1']
    penzynes = ['C1=C[NH]C=C1', 'C1=C[NH]C=N1', 'C1=C[NH]N=C1', 'C1=C[NH]N=N1', 'C1=COC=C1', 'C1=COC=N1', 'C1=CON=C1', 'C1=CSC=C1', 'C1=CSC=N1', 'C1=CSN=C1', 'C1=CSN=N1', 'C1=NN=C[NH]1', 'C1=NN=CO1', 'C1=NN=CS1', 'C1=N[NH]C=N1', 'C1=N[NH]N=C1', 'C1=N[NH]N=N1', 'C1=NN=N[NH]1', 'C1=NN=NS1', 'C1=NOC=N1', 'C1=NON=C1', 'C1=NSC=N1', 'C1=NSN=C1']

    def __init__(self, smiles_list):
        self.vocab = smiles_list
        self.vmap = {x:i for i,x in enumerate(self.vocab)}
        self.slots = [get_slots(smiles) for smiles in self.vocab]
        Vocab.benzynes = [s for s in smiles_list if s.count('=') >= 2 and Chem.MolFromSmiles(s).GetNumAtoms() == 6] + ['C1=CCNCC1']
        Vocab.penzynes = [s for s in smiles_list if s.count('=') >= 2 and Chem.MolFromSmiles(s).GetNumAtoms() == 5] + ['C1=NCCN1','C1=NNCC1']
        
    def get_index(self, smiles):
        return self.vmap[smiles]

    def get_smiles(self, idx):
        return self.vocab[idx]

    def get_slots(self, idx):
        return copy.deepcopy(self.slots[idx])

    def size(self):
        return len(self.vocab)




class SmilesCharDictionary(object):
    """
    A fixed dictionary for druglike SMILES.
    Enables smile<->token conversion.

    With a space:0 for padding, Q:1 as the start token and end_of_line \n:2 as the stop token.
    """

    PAD = ' '
    BEGIN = 'Q'
    END = '\n'

    def __init__(self) -> None:
        self.forbidden_symbols = {'Ag', 'Al', 'Am', 'Ar', 'At', 'Au', 'D', 'E', 'Fe', 'G', 'K', 'L', 'M', 'Ra', 'Re',
                                  'Rf', 'Rg', 'Rh', 'Ru', 'T', 'U', 'V', 'W', 'Xe',
                                  'Y', 'Zr', 'a', 'd', 'f', 'g', 'h', 'k', 'm', 'si', 't', 'te', 'u', 'v', 'y'}

        self.char_idx = {self.PAD: 0, self.BEGIN: 1, self.END: 2, '#': 20, '%': 22, '(': 25, ')': 24, '+': 26, '-': 27,
                         '.': 30,
                        #  '*': 30,
                         '*': 47,
                         '0': 32, '1': 31, '2': 34, '3': 33, '4': 36, '5': 35, '6': 38, '7': 37, '8': 40,
                         '9': 39, '=': 41, 'A': 7, 'B': 11, 'C': 19, 'F': 4, 'H': 6, 'I': 5, 'N': 10,
                         'O': 9, 'P': 12, 'S': 13, 'X': 15, 'Y': 14, 'Z': 3, '[': 16, ']': 18,
                         'b': 21, 'c': 8, 'n': 17, 'o': 29, 'p': 23, 's': 28,
                         "@": 42, "R": 43, '/': 44, "\\": 45, 'E': 46, 
                         }

        self.idx_char = {v: k for k, v in self.char_idx.items()}
        self.encode_dict = {"Br": 'Y', "Cl": 'X', "Si": 'A', 'Se': 'Z', '@@': 'R', 'se': 'E'}
        self.decode_dict = {v: k for k, v in self.encode_dict.items()}

    def allowed(self, smiles) -> bool:
        """
        Determine if smiles string has illegal symbols

        Args:
            smiles: SMILES string

        Returns:
            True if all legal
        """
        for symbol in self.forbidden_symbols:
            if symbol in smiles:
                # print('Forbidden symbol {:<2}  in  {}'.format(symbol, smiles))
                return False
        return True

    def encode(self, smiles: str) -> str:
        """
        Replace multi-char tokens with single tokens in SMILES string.

        Args:
            smiles: SMILES string

        Returns:
            sanitized SMILE string with only single-char tokens
        """

        temp_smiles = smiles
        for symbol, token in self.encode_dict.items():
            temp_smiles = temp_smiles.replace(symbol, token)
        return temp_smiles

    def decode(self, smiles):
        """
        Replace special tokens with their multi-character equivalents.

        Args:
            smiles: SMILES string

        Returns:
            SMILES string with possibly multi-char tokens
        """
        temp_smiles = smiles
        for symbol, token in self.decode_dict.items():
            temp_smiles = temp_smiles.replace(symbol, token)
        return temp_smiles

    def get_char_num(self) -> int:
        """
        Returns:
            number of characters in the alphabet
        """
        return len(self.idx_char)

    @property
    def begin_idx(self) -> int:
        return self.char_idx[self.BEGIN]

    @property
    def end_idx(self) -> int:
        return self.char_idx[self.END]

    @property
    def pad_idx(self) -> int:
        return self.char_idx[self.PAD]

    def matrix_to_smiles(self, array):
        """
        Converts an matrix of indices into their SMILES representations

        Args:
            array: torch tensor of indices, one molecule per row

        Returns: a list of SMILES, without the termination symbol
        """
        smiles_strings = []

        for row in array:
            predicted_chars = []

            for j in row:
                next_char = self.idx_char[j.item()]
                if next_char == self.END:
                    break
                predicted_chars.append(next_char)

            smi = ''.join(predicted_chars)
            smi = self.decode(smi)
            smiles_strings.append(smi)

        return smiles_strings



import torch
import copy
import random
import networkx as nx
import numpy as np

from torch_geometric.data import Data

def cumsum(self, key, item):
        r"""If :obj:`True`, the attribute :obj:`key` with content :obj:`item`
        should be added up cumulatively before concatenated together.
        .. note::
            This method is for internal use only, and should only be overridden
            if the batch concatenation process is corrupted for a specific data
            attribute.
        """
        return key in ["edge_index", "edge_index_substruct", "edge_index_context", "overlap_context_substruct_idx", "center_substruct_idx"]

class ExtractSubstructureContextPair:
    def __init__(self, k, l1, l2):
        """
        Randomly selects a node from the data object, and adds attributes
        that contain the substructure that corresponds to k hop neighbours
        rooted at the node, and the context substructures that corresponds to
        the subgraph that is between l1 and l2 hops away from the
        root node.
        :param k:
        :param l1:
        :param l2:
        """
        self.k = k
        self.l1 = l1
        self.l2 = l2

        # for the special case of 0, addresses the quirk with
        # single_source_shortest_path_length
        if self.k == 0:
            self.k = -1
        if self.l1 == 0:
            self.l1 = -1
        if self.l2 == 0:
            self.l2 = -1

    def __call__(self, data, root_idx=None):
        """

        :param data: pytorch geometric data object
        :param root_idx: If None, then randomly samples an atom idx.
        Otherwise sets atom idx of root (for debugging only)
        :return: None. Creates new attributes in original data object:
        data.center_substruct_idx
        data.x_substruct
        data.edge_attr_substruct
        data.edge_index_substruct
        data.x_context
        data.edge_attr_context
        data.edge_index_context
        data.overlap_context_substruct_idx
        """
        num_atoms = data.x.size()[0]
        if root_idx == None:
            root_idx = random.sample(range(num_atoms), 1)[0]

        G = graph_data_obj_to_nx_simple(data)  # same ordering as input data obj

        # Get k-hop subgraph rooted at specified atom idx
        substruct_node_idxes = nx.single_source_shortest_path_length(G,
                                                                     root_idx,
                                                                     self.k).keys()
        if len(substruct_node_idxes) > 0:
            substruct_G = G.subgraph(substruct_node_idxes)
            substruct_G, substruct_node_map = reset_idxes(substruct_G)  # need
            # to reset node idx to 0 -> num_nodes - 1, otherwise data obj does not
            # make sense, since the node indices in data obj must start at 0
            substruct_data = nx_to_graph_data_obj_simple(substruct_G)
            data.substruct_node_idxes = torch.tensor(list(substruct_node_idxes))
            data.x_substruct = substruct_data.x
            data.edge_attr_substruct = substruct_data.edge_attr
            data.edge_index_substruct = substruct_data.edge_index
            data.center_substruct_idx = torch.tensor([substruct_node_map[
                                                          root_idx]])  # need
            # to convert center idx from original graph node ordering to the
            # new substruct node ordering

        # Get subgraphs that is between l1 and l2 hops away from the root node
        l1_node_idxes = nx.single_source_shortest_path_length(G, root_idx,
                                                              self.l1).keys()
        l2_node_idxes = nx.single_source_shortest_path_length(G, root_idx,
                                                              self.l2).keys()
        context_node_idxes = set(l1_node_idxes).symmetric_difference(
            set(l2_node_idxes))
        if len(context_node_idxes) > 0:
            context_G = G.subgraph(context_node_idxes)
            context_G, context_node_map = reset_idxes(context_G)  # need to
            # reset node idx to 0 -> num_nodes - 1, otherwise data obj does not
            # make sense, since the node indices in data obj must start at 0
            context_data = nx_to_graph_data_obj_simple(context_G)
            data.context_node_idxes = torch.tensor(list(context_node_idxes))
            data.x_context = context_data.x
            data.edge_attr_context = context_data.edge_attr
            data.edge_index_context = context_data.edge_index

        # Get indices of overlapping nodes between substruct and context,
        # WRT context ordering
        context_substruct_overlap_idxes = list(set(
            context_node_idxes).intersection(set(substruct_node_idxes)))
        if len(context_substruct_overlap_idxes) > 0:
            context_substruct_overlap_idxes_reorder = [context_node_map[old_idx]
                                                       for
                                                       old_idx in
                                                       context_substruct_overlap_idxes]
            # need to convert the overlap node idxes, which is from the
            # original graph node ordering to the new context node ordering
            data.overlap_context_substruct_idx = \
                torch.tensor(context_substruct_overlap_idxes_reorder)

        return data
    
def reset_idxes(G):
    """
    Resets node indices such that they are numbered from 0 to num_nodes - 1
    :param G:
    :return: copy of G with relabelled node indices, mapping
    """
    mapping = {}
    for new_idx, old_idx in enumerate(G.nodes()):
        mapping[old_idx] = new_idx
    new_G = nx.relabel_nodes(G, mapping, copy=True)
    return new_G, mapping

def graph_data_obj_to_nx_simple(data):
    """
    Converts graph Data object required by the pytorch geometric package to
    network x data object. NB: Uses simplified atom and bond features,
    and represent as indices. NB: possible issues with recapitulating relative
    stereochemistry since the edges in the nx object are unordered.
    :param data: pytorch geometric Data object
    :return: network x object
    """
    G = nx.Graph()

    # atoms
    atom_features = data.x.cpu().numpy()
    num_atoms = atom_features.shape[0]
    for i in range(num_atoms):
        atomic_num_idx, chirality_tag_idx, _, _, _, _, _, _, _ = atom_features[i]
        G.add_node(i, atom_num_idx=atomic_num_idx, chirality_tag_idx=chirality_tag_idx)
        pass

    # bonds
    edge_index = data.edge_index.cpu().numpy()
    edge_attr = data.edge_attr.cpu().numpy()
    num_bonds = edge_index.shape[1]
    for j in range(0, num_bonds, 2):
        begin_idx = int(edge_index[0, j])
        end_idx = int(edge_index[1, j])
        bond_type_idx, bond_dir_idx, _ = edge_attr[j]
        if not G.has_edge(begin_idx, end_idx):
            G.add_edge(begin_idx, end_idx, bond_type_idx=bond_type_idx,
                       bond_dir_idx=bond_dir_idx)

    return G

def nx_to_graph_data_obj_simple(G):
    """
    Converts nx graph to pytorch geometric Data object. Assume node indices
    are numbered from 0 to num_nodes - 1. NB: Uses simplified atom and bond
    features, and represent as indices. NB: possible issues with
    recapitulating relative stereochemistry since the edges in the nx
    object are unordered.
    :param G: nx graph obj
    :return: pytorch geometric Data object
    """
    # atoms
    num_atom_features = 2  # atom type,  chirality tag
    atom_features_list = []
    for _, node in G.nodes(data=True):
        atom_feature = [node['atom_num_idx'], node['chirality_tag_idx']]
        atom_features_list.append(atom_feature)
    x = torch.tensor(np.array(atom_features_list), dtype=torch.long)

    # bonds
    num_bond_features = 2  # bond type, bond direction
    if len(G.edges()) > 0:  # mol has bonds
        edges_list = []
        edge_features_list = []
        for i, j, edge in G.edges(data=True):
            edge_feature = [edge['bond_type_idx'], edge['bond_dir_idx']]
            edges_list.append((i, j))
            edge_features_list.append(edge_feature)
            edges_list.append((j, i))
            edge_features_list.append(edge_feature)

        # data.edge_index: Graph connectivity in COO format with shape [2, num_edges]
        edge_index = torch.tensor(np.array(edges_list).T, dtype=torch.long)

        # data.edge_attr: Edge feature matrix with shape [num_edges, num_edge_features]
        edge_attr = torch.tensor(np.array(edge_features_list),
                                 dtype=torch.long)
    else:   # mol has no bonds
        edge_index = torch.empty((2, 0), dtype=torch.long)
        edge_attr = torch.empty((0, num_bond_features), dtype=torch.long)

    data = Data(x=x, edge_index=edge_index, edge_attr=edge_attr)

    return data


from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Optional, Dict, List, Type, Any, ClassVar, Union, Tuple, Callable, Literal
import torch
import os
import numpy as np
from ..utils.checkpoint import LocalCheckpointManager, HuggingFaceCheckpointManager
from ..utils.checker import MolecularInputChecker

@dataclass
class BaseModel(ABC):
    """Base class for molecular models with shared functionality.
    
    This abstract class provides common methods and utilities for molecular models,
    including model initialization, saving/loading, and parameter management.
    """
    
    device: Optional[torch.device] = field(default=None)
    model_name: str = field(default="BaseModel")
    model_class: Optional[Type[torch.nn.Module]] = field(default=None, init=False) # used for model initialization
    model: Optional[torch.nn.Module] = field(default=None, init=False) # initialized model
    is_fitted_: bool = field(default=False, init=False)

    def __post_init__(self):
        """Initialize common device settings after instance creation.
        
        Sets the device to CUDA if available, otherwise CPU, when no device is specified.
        Converts string device specifications to torch.device objects.
        """
        if self.device is None:
            self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        elif isinstance(self.device, str):
            self.device = torch.device(self.device)

    @abstractmethod
    def _setup_optimizers(self) -> Tuple[torch.optim.Optimizer, Optional[Any]]:
        """Set up optimizers for model training.
        
        Returns
        -------
        Tuple[torch.optim.Optimizer, Optional[Any]]
            Tuple containing the primary optimizer and an optional secondary optimizer or scheduler.
        """
        pass
    
    @abstractmethod
    def _train_epoch(self, train_loader, optimizer):
        """Train the model for one epoch.
        
        Parameters
        ----------
        train_loader : torch.utils.data.DataLoader
            DataLoader containing training batches
        optimizer : torch.optim.Optimizer
            Optimizer to use for parameter updates
            
        Returns
        -------
        dict
            Training metrics for the epoch
        """
        pass

    @abstractmethod
    def _get_model_params(self, checkpoint: Optional[Dict] = None) -> Dict[str, Any]:
        """Get model parameters used for model initialization.
        
        Parameters
        ----------
        checkpoint : Optional[Dict], default=None
            Optional dictionary containing model checkpoint data
            
        Returns
        -------
        Dict[str, Any]
            Dictionary of parameters to initialize the model
        """
        pass

    @staticmethod
    def _get_param_names(self) -> List[str]:
        """Get parameter names in the modeling class.
        
        Returns
        -------
        List[str]
            List of parameter names that can be configured
        """
        # return ["model_name", "model_class", "is_fitted_"]
        return ["model_name", "is_fitted_"]

    def get_params(self, deep: bool = True) -> Dict[str, Any]:
        """Get parameters for this estimator.
        
        Parameters
        ----------
        deep : bool, default=True
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.
                 
        Returns
        -------
        Dict[str, Any]
            Dictionary of parameter names mapped to their values
        """
        out = {}
        for key in self._get_param_names():
            value = getattr(self, key)
            if deep and hasattr(value, "get_params"):
                deep_items = value.get_params().items()
                out.update((key + "__" + k, val) for k, val in deep_items)
            out[key] = value
        return out

    def set_params(self, **params) -> "BaseModel":
        """Set parameters for this estimator.
        
        Parameters
        ----------
        **params
            Parameter names mapped to their values
            
        Returns
        -------
        BaseModel
            Self instance for method chaining
            
        Raises
        ------
        ValueError
            If an invalid parameter is provided
        """
        valid_params = self.get_params(deep=True)
        for key, value in params.items():
            if key not in valid_params:
                raise ValueError(f"Invalid parameter {key} for model {self}")
            setattr(self, key, value)
        return self

    def _initialize_model(
        self,
        model_class: Type[torch.nn.Module],
        checkpoint: Optional[Dict] = None
    ) -> torch.nn.Module:
        """Initialize the model with parameters or a checkpoint.
        
        Parameters
        ----------
        model_class : Type[torch.nn.Module]
            PyTorch module class to instantiate
        checkpoint : Optional[Dict], default=None
            Optional dictionary containing model checkpoint data
            
        Returns
        -------
        torch.nn.Module
            Initialized PyTorch model
        """
        model_params = self._get_model_params(checkpoint)
        self.model = model_class(**model_params)
        self.model = self.model.to(self.device)
        
        if checkpoint is not None:
            self.model.load_state_dict(checkpoint["model_state_dict"])
        return self.model
    
    def _validate_inputs(
        self, X: List[str], y: Optional[Union[List, np.ndarray]] = None, num_task: int = 0, num_pretask: int = 0, return_rdkit_mol: bool = True
    ) -> Tuple[Union[List[str], List["Chem.Mol"]], Optional[np.ndarray]]:
        """Validate molecular inputs and targets.
        
        Parameters
        ----------
        X : List[str]
            List of SMILES strings representing molecules
        y : Optional[Union[List, np.ndarray]], default=None
            Optional target values for supervised learning
        num_task : int, default=0
            Number of prediction tasks
        num_pretask : int, default=0
            Number of pre-training tasks
        return_rdkit_mol : bool, default=True
            Whether to return RDKit Mol objects instead of SMILES
            
        Returns
        -------
        Tuple[Union[List[str], List["Chem.Mol"]], Optional[np.ndarray]]
            Tuple of validated inputs and targets
        """
        return MolecularInputChecker.validate_inputs(X, y, num_task, num_pretask, return_rdkit_mol)

    def save_to_local(self, path: str) -> None:
        """Save model to local disk.
        
        Parameters
        ----------
        path : str
            File path to save the model
            
        Raises
        ------
        ValueError
            If the model is not fitted
        """
        if not self.is_fitted_:
            raise ValueError("Model must be fitted before saving to local disk.")
        LocalCheckpointManager.save_model_to_local(self, path)

    def load_from_local(self, path: str) -> None:
        """Load model from local disk.
        
        Parameters
        ----------
        path : str
            File path to load the model from
        """
        LocalCheckpointManager.load_model_from_local(self, path)

    def save_to_hf(
        self,
        repo_id: str,
        task_id: str = "default",
        metadata_dict: Optional[Dict[str, Any]] = None,
        metrics: Optional[Dict[str, float]] = None,
        commit_message: str = "Update model",
        hf_token: Optional[str] = None,
        private: bool = False,
        config_filename: Optional[str] = 'config.json',
    ) -> None:
        """Save model to Hugging Face Hub.
        
        Parameters
        ----------
        repo_id : str
            Hugging Face repository ID
        task_id : str, default="default"
            Task identifier for the model
        metadata_dict : Optional[Dict[str, Any]], default=None
            Optional metadata to store with the model
        metrics : Optional[Dict[str, float]], default=None
            Optional performance metrics to store with the model
        commit_message : str, default="Update model"
            Git commit message
        hf_token : Optional[str], default=None
            Hugging Face authentication token
        private : bool, default=False
            Whether the repository should be private
        config_filename : Optional[str], default='config.json'
            Name of the configuration file to save to the repository
        Raises
        ------
        ValueError
            If the model is not fitted
        """
        if not self.is_fitted_:
            raise ValueError("Model must be fitted before saving to Hugging Face Hub.")
        HuggingFaceCheckpointManager.push_to_huggingface(
            model_instance=self,
            repo_id=repo_id,
            task_id=task_id,
            metadata_dict=metadata_dict,
            metrics=metrics,
            commit_message=commit_message,
            token=hf_token,
            private=private,
            config_filename=config_filename,
        )

    def load_from_hf(self, repo_id: str, local_cache: Optional[str] = None, config_filename: Optional[str] = 'config.json') -> None:
        """Load model from Hugging Face Hub.
        
        Parameters
        ----------
        repo_id : str
            Hugging Face repository ID
        local_cache : str, default=None
            Local path to save the model
        config_filename : str, default='config.json'
            Name of the configuration file to load from the repository
        """
        if local_cache is None:
            local_cache = 'model.pt'
        HuggingFaceCheckpointManager.load_model_from_hf(self, repo_id, local_cache, config_filename=config_filename)

    def save(self, path: Optional[str] = None, repo_id: Optional[str] = None, **kwargs) -> None:
        """Automatic save to either local disk or Hugging Face Hub.
        
        Parameters
        ----------
        path : Optional[str], default=None
            File path for local saving (required if repo_id is None)
        repo_id : Optional[str], default=None
            Hugging Face repository ID for remote saving
        **kwargs
            Additional arguments passed to save_to_hf
            
        Raises
        ------
        ValueError
            If path is None when repo_id is None
        """
        # if both path and repo_id are None, raise an error
        if path is None and repo_id is None:
            raise ValueError("path must be provided if repo_id is not given.")
        
        if repo_id is not None:
            self.save_to_hf(repo_id=repo_id, **kwargs)

        if path is not None:
            self.save_to_local(path)

    def load(self, path: Optional[str] = None, repo_id: Optional[str] = None, **kwargs) -> None:
        """Automatic load from either local disk or Hugging Face Hub.
        
        Parameters
        ----------
        path : Optional[str], default=None
            File path for local loading.
        repo_id : Optional[str], default=None
            Hugging Face repository ID for remote loading. If path is provided, repo_id is ignored.
        **kwargs
            Additional arguments passed to load_from_hf
            
        Raises
        ------
        FileNotFoundError
            If no local file is found and no repo_id is provided
        """
        if path is not None:
            if os.path.exists(path):
                self.load_from_local(path)
            else:
                raise FileNotFoundError(f"No local file found at '{path}'.")
        else:
            if repo_id is None:
                raise ValueError("repo_id must be provided if path is not given.")
            self.load_from_hf(repo_id, **kwargs)

    def _check_is_fitted(self) -> None:
        """Check if the model is fitted.
        
        Raises
        ------
        AttributeError
            If the model is not fitted
        """
        if not self.is_fitted_:
            raise AttributeError("This model is not fitted yet. Call 'fit' before using it.")

    def __str__(self, N_CHAR_MAX: int = 700) -> str:
        """Return a string representation of the model.
        
        Parameters
        ----------
        N_CHAR_MAX : int, default=700
            Maximum number of characters in the string representation
            
        Returns
        -------
        str
            String representation of the model
        """
        attributes = {
            name: value
            for name, value in sorted(self.__dict__.items())
            if not name.startswith("_") and not callable(value)
        }
        attributes = {k: v for k, v in attributes.items() if k != "fitting_loss"}

        def format_value(v):
            """Helper to format values for representation."""
            if isinstance(v, (float, np.float32, np.float64)):
                return f"{v:.3g}"
            elif isinstance(v, (list, tuple, np.ndarray)) and len(v) > 6:
                return f"{v[:3]}...{v[-3:]}"
            elif isinstance(v, str) and len(v) > 50:
                return f"'{v[:25]}...{v[-22:]}'"
            elif isinstance(v, dict) and len(v) > 6:
                return f"{{{', '.join(f'{k}: {v}' for k, v in list(v.items())[:3])}...}}"
            elif isinstance(v, torch.nn.Module):
                return f"{v.__class__.__name__}(...)"
            return repr(v)

        class_name = self.__class__.__name__
        important_attrs = ["model_name", "is_fitted_", "task_type", "num_task"]
        attributes_str = [f"{attr}={format_value(attributes.pop(attr))}" for attr in important_attrs if attr in attributes]
        attributes_str += [f"{k}={format_value(v)}" for k, v in sorted(attributes.items())]

        content = ",\n    ".join(attributes_str)
        repr_str = f"{class_name}(\n    {content}\n)"
        
        if len(repr_str) > N_CHAR_MAX:
            repr_str = "\n".join([repr_str[:N_CHAR_MAX//2], "...", repr_str[-N_CHAR_MAX//2:]])
        
        return repr_str
    


import torch
import torch.nn as nn
import numpy as np
from torch_geometric.nn import global_add_pool, global_mean_pool, global_max_pool

from .utils import split_batch, relabel, set_masks, clear_masks
from ...nn import GNN_node, GNN_node_Virtualnode
from ...utils import init_weights

class DIR(nn.Module):
    def __init__(
        self,
        num_task,
        num_layer,
        causal_ratio=0.8,
        hidden_size=300,
        gnn_type="gin-virtual",
        drop_ratio=0.5,
        norm_layer="batch_norm",
        graph_pooling="sum",
        augmented_feature=['maccs', 'morgan']
    ):
        super(DIR, self).__init__()
        gnn_name = gnn_type.split("-")[0]
        self.num_task = num_task
        self.hidden_size = hidden_size
        self.causal_ratio = causal_ratio
        if "virtual" in gnn_type:
            self.graph_encoder = GNN_node_Virtualnode(
                num_layer,
                hidden_size,
                JK="last",
                drop_ratio=drop_ratio,
                residual=True,
                gnn_name=gnn_name,
                norm_layer=norm_layer,
                encode_atom=False
            )
        else:
            self.graph_encoder = GNN_node(
                num_layer,
                hidden_size,
                JK="last",
                drop_ratio=drop_ratio,
                residual=True,
                gnn_name=gnn_name,
                norm_layer=norm_layer,
                encode_atom=False
            )

        self.causal_att_net = CausalAttNet(causal_ratio=causal_ratio, hidden_size=hidden_size)
        
        graph_dim = hidden_size
        self.augmented_feature = augmented_feature
        if graph_pooling == "sum":
            self.pool = global_add_pool
        elif graph_pooling == "mean":
            self.pool = global_mean_pool
        elif graph_pooling == "max":
            self.pool = global_max_pool
        else:
            raise ValueError(f"Invalid graph pooling type {graph_pooling}.")
        
        if augmented_feature:
            if "morgan" in augmented_feature:
                graph_dim += 1024
            if "maccs" in augmented_feature:
                graph_dim += 167
        
        self.causal_lin = torch.nn.Linear(graph_dim, num_task)
        self.conf_lin = torch.nn.Linear(hidden_size, num_task)
    
    def get_graph_rep(self, x, edge_index, edge_attr, batch):
        h_node, _ = self.graph_encoder(x, edge_index, edge_attr, batch)
        h_graph = self.pool(h_node, batch)
        return h_graph
    
    def get_causal_pred(self, batched_data, h_graph):
        h_graph = self._augment_graph_features(batched_data, h_graph)
        return self.causal_lin(h_graph)
    
    def get_conf_pred(self, conf_graph_x):
        return self.conf_lin(conf_graph_x)
    
    def get_comb_pred(self, batched_data, causal_graph_x, conf_graph_x):
        causal_graph_x = self._augment_graph_features(batched_data, causal_graph_x)
        causal_pred = self.causal_lin(causal_graph_x)
        conf_pred = self.conf_lin(conf_graph_x).detach()
        return torch.sigmoid(conf_pred) * causal_pred

    def initialize_parameters(self, seed=None):
        if seed is not None:
            torch.manual_seed(seed)
        
        # Initialize the main components
        init_weights(self.graph_encoder)
        init_weights(self.causal_att_net)
        init_weights(self.causal_lin)
        init_weights(self.conf_lin)
        
        # Reset all parameters using PyTorch Geometric's reset function
        def reset_parameters(module):
            if hasattr(module, 'reset_parameters'):
                module.reset_parameters()
            elif hasattr(module, 'weight') and hasattr(module.weight, 'data'):
                init_weights(module)
        
        self.apply(reset_parameters)

    def _augment_graph_features(self, batched_data, h):
        if self.augmented_feature:
            if 'morgan' in self.augmented_feature:
                morgan = batched_data.morgan.type_as(h)
                h = torch.cat((h, morgan), dim=1)
            if 'maccs' in self.augmented_feature:
                maccs = batched_data.maccs.type_as(h)
                h = torch.cat((h, maccs), dim=1)
        return h
    
    def compute_loss(self, batched_data, criterion, alpha_prime):
        (causal_x, causal_edge_index, causal_edge_attr, causal_edge_weight, causal_batch),\
        (conf_x, conf_edge_index, conf_edge_attr, conf_edge_weight, conf_batch), pred_edge_weight = self.causal_att_net(batched_data)
            
        set_masks(causal_edge_weight, self.graph_encoder)
        set_masks(causal_edge_weight, self.causal_lin)
        set_masks(causal_edge_weight, self.conf_lin)
        causal_rep = self.get_graph_rep(x=causal_x, edge_index=causal_edge_index, edge_attr=causal_edge_attr, batch=causal_batch)
        causal_out = self.get_causal_pred(batched_data, causal_rep)
        clear_masks(self.graph_encoder)
        clear_masks(self.causal_lin)
        clear_masks(self.conf_lin)

        set_masks(conf_edge_weight, self.graph_encoder)
        set_masks(conf_edge_weight, self.causal_lin)
        set_masks(conf_edge_weight, self.conf_lin)
        conf_rep = self.get_graph_rep(x=conf_x, edge_index=conf_edge_index, edge_attr=conf_edge_attr, batch=conf_batch).detach()
        conf_out = self.get_conf_pred(conf_rep)
        clear_masks(self.graph_encoder)
        clear_masks(self.causal_lin)
        clear_masks(self.conf_lin)

        batched_data.y = batched_data.y[0:causal_out.size(0),:]
        is_labeled = batched_data.y == batched_data.y
        
        causal_loss = criterion(
            causal_out.to(torch.float32)[is_labeled], 
            batched_data.y.to(torch.float32)[is_labeled]
            ).mean()
        conf_loss = criterion(
            conf_out.to(torch.float32)[is_labeled], 
            batched_data.y.to(torch.float32)[is_labeled]
            ).mean()

        env_loss = torch.tensor([]).to(batched_data.y.device)
        for conf in conf_rep:
            rep_out = self.get_comb_pred(batched_data, causal_rep, conf)
            tmp = criterion(rep_out.to(torch.float32)[is_labeled], batched_data.y.to(torch.float32)[is_labeled]).mean()
            env_loss = torch.cat([env_loss, tmp.unsqueeze(0)])
        causal_loss += alpha_prime * env_loss.mean()
        env_loss = alpha_prime * torch.var(env_loss * conf_rep.size(0))
        
        return causal_loss, conf_loss, env_loss

    def forward(self, batched_data):
        (causal_x, causal_edge_index, causal_edge_attr, causal_edge_weight, causal_batch), _, _ = self.causal_att_net(batched_data)
        set_masks(causal_edge_weight, self.graph_encoder)
        set_masks(causal_edge_weight, self.causal_lin)
        set_masks(causal_edge_weight, self.conf_lin)

        causal_rep = self.get_graph_rep(x=causal_x, edge_index=causal_edge_index, edge_attr=causal_edge_attr, batch=causal_batch)
        causal_out = self.get_causal_pred(batched_data, causal_rep)
        clear_masks(self.graph_encoder)
        clear_masks(self.causal_lin)
        clear_masks(self.conf_lin)

        return {"prediction": causal_out, "representation": causal_rep}

    @staticmethod
    def _disable_batchnorm_tracking(model):
        def fn(module):
            if isinstance(module, torch.nn.modules.batchnorm._BatchNorm):
                module.track_running_stats = False

        model.apply(fn)
    
    @staticmethod
    def _enable_batchnorm_tracking(model):
        def fn(module):
            if isinstance(module, torch.nn.modules.batchnorm._BatchNorm):
                module.track_running_stats = True

        model.apply(fn)

class CausalAttNet(nn.Module):
    def __init__(self, causal_ratio, hidden_size):
        super(CausalAttNet, self).__init__()
        self.gnn_node = GNN_node_Virtualnode(2, hidden_size, JK="last", drop_ratio=0., residual=True, gnn_name="gin", norm_layer="batch_norm")
        self.linear = nn.Linear(hidden_size*2, 1)
        self.ratio = causal_ratio
        
    def forward(self, data):
        x, _ = self.gnn_node(data.x, data.edge_index, data.edge_attr, data.batch)

        row, col = data.edge_index
        edge_rep = torch.cat([x[row], x[col]], dim=-1)
        pred_edge_weight = self.linear(edge_rep).view(-1)
        causal_edge_index = torch.LongTensor([[],[]]).to(x.device)
        causal_edge_weight = torch.tensor([]).to(x.device)
        causal_edge_attr = torch.LongTensor([]).to(x.device)
        conf_edge_index = torch.LongTensor([[],[]]).to(x.device)
        conf_edge_weight = torch.tensor([]).to(x.device)
        conf_edge_attr = torch.LongTensor([]).to(x.device)

        edge_indices, _, _, num_edges, cum_edges = split_batch(data)
        for edge_index, N, C in zip(edge_indices, num_edges, cum_edges):
            n_reserve =  int(self.ratio * N)
            edge_attr = data.edge_attr[C:C+N]
            single_mask = pred_edge_weight[C:C+N]
            single_mask_detach = pred_edge_weight[C:C+N].detach().cpu().numpy()
            rank = np.argpartition(-single_mask_detach, n_reserve)
            idx_reserve, idx_drop = rank[:n_reserve], rank[n_reserve:]

            causal_edge_index = torch.cat([causal_edge_index, edge_index[:, idx_reserve]], dim=1)
            conf_edge_index = torch.cat([conf_edge_index, edge_index[:, idx_drop]], dim=1)
            causal_edge_weight = torch.cat([causal_edge_weight, single_mask[idx_reserve]])
            conf_edge_weight = torch.cat([conf_edge_weight, -1 * single_mask[idx_drop]])
            causal_edge_attr = torch.cat([causal_edge_attr, edge_attr[idx_reserve]])
            conf_edge_attr = torch.cat([conf_edge_attr, edge_attr[idx_drop]])

        causal_x, causal_edge_index, causal_batch, _ = relabel(x, causal_edge_index, data.batch)
        conf_x, conf_edge_index, conf_batch, _ = relabel(x, conf_edge_index, data.batch)

        return (causal_x, causal_edge_index, causal_edge_attr, causal_edge_weight, causal_batch),\
                (conf_x, conf_edge_index, conf_edge_attr, conf_edge_weight, conf_batch),\
                pred_edge_weight



import numpy as np
from tqdm import tqdm
from typing import Optional, Union, Dict, Any, Tuple, List, Literal, Type
from dataclasses import dataclass, field

import torch
from torch_geometric.loader import DataLoader
from torch_geometric.data import Data

from .model import GNN
from ..constant import GNN_ENCODER_MODELS, GNN_ENCODER_READOUTS, GNN_ENCODER_PARAMS
from ...base import BaseMolecularEncoder
from ...utils import graph_from_smiles
from ...utils import PSEUDOTASK

ALLOWABLE_ENCODER_MODELS = GNN_ENCODER_MODELS
ALLOWABLE_ENCODER_READOUTS = GNN_ENCODER_READOUTS

@dataclass
class SupervisedMolecularEncoder(BaseMolecularEncoder):
    """This encoder implements a GNN model for supervised molecular representation learning with user-defined or predefined fingerprint/calculated property tasks.

    Parameters
    ----------
    num_task : int, optional
        Number of user-defined tasks for supervised pretraining. If it is specified, user must provide y_train in the fit function.
    predefined_task : List[str], optional
        List of predefined tasks to use. Must be from the supported task list ["morgan", "maccs", "logP"]. If None and num_task is None, all predefined tasks will be used.
    encoder_type : str, default="gin-virtual" 
        Type of GNN architecture to use. One of ["gin-virtual", "gcn-virtual", "gin", "gcn"].
    readout : str, default="sum" 
        Method for aggregating node features to obtain graph-level representations. One of ["sum", "mean", "max"].
    num_layer : int, default=5
        Number of GNN layers.
    hidden_size : int, default=300
        Dimension of hidden node features.
    drop_ratio : float, default=0.5
        Dropout probability.
    norm_layer : str, default="batch_norm"
        Type of normalization layer to use. One of ["batch_norm", "layer_norm", "instance_norm", "graph_norm", "size_norm", "pair_norm"].
    batch_size : int, default=128
        Number of samples per batch for training.
    epochs : int, default=500
        Maximum number of training epochs.
    learning_rate : float, default=0.001
        Learning rate for optimizer.
    grad_clip_value : float, optional
        Maximum norm of gradients for gradient clipping.
    weight_decay : float, default=0.0
        L2 regularization strength.
    use_lr_scheduler : bool, default=False
        Whether to use a learning rate scheduler.
    scheduler_factor : float, default=0.5
        Factor by which to reduce the learning rate when plateau is detected.
    scheduler_patience : int, default=5
        Number of epochs with no improvement after which learning rate will be reduced.
    verbose : bool, default=False
        Whether to print progress information during training.
    """
    # pretraining task
    num_task: Optional[int] = None
    predefined_task: Optional[List[str]] = None
    # Model parameters    
    encoder_type: str = "gin-virtual"
    readout: str = "sum"
    num_layer: int = 5
    hidden_size: int = 300
    drop_ratio: float = 0.5
    norm_layer: str = "batch_norm"
    
    # Training parameters
    batch_size: int = 128
    epochs: int = 500
    learning_rate: float = 0.001
    grad_clip_value: Optional[float] = None
    weight_decay: float = 0.0
    
    # Scheduler parameters
    use_lr_scheduler: bool = False
    scheduler_factor: float = 0.5
    scheduler_patience: int = 5
    
    # Other parameters
    verbose: bool = False
    model_name: str = "SupervisedMolecularEncoder"
    
    # Non-init fields
    fitting_loss: List[float] = field(default_factory=list, init=False)
    fitting_epoch: int = field(default=0, init=False)
    model_class: Type[GNN] = field(default=GNN, init=False)
    
    def __post_init__(self):
        """Initialize the model after dataclass initialization."""
        super().__post_init__()
        self.num_pretask = None

        if self.encoder_type not in ALLOWABLE_ENCODER_MODELS:
            raise ValueError(f"Invalid encoder: {self.encoder_type}. Currently only {ALLOWABLE_ENCODER_MODELS} are supported.")
        if self.readout not in ALLOWABLE_ENCODER_READOUTS:
            raise ValueError(f"Invalid readout: {self.readout}. Currently only {ALLOWABLE_ENCODER_READOUTS} are supported.")
        if self.predefined_task is not None:
            for task in self.predefined_task:
                if task not in PSEUDOTASK.keys():
                    raise ValueError(f"Invalid predefined_task: {task}. Currently only {PSEUDOTASK.keys()} are supported.")
        
        # Calculate number of predefined tasks if any are specified
        num_pretask = 0
        if self.predefined_task is not None:
            num_pretask = sum(PSEUDOTASK[task][0] for task in self.predefined_task)
        elif self.predefined_task is None and self.num_task is None:
            # Use all predefined tasks if none specified
            self.predefined_task = list(PSEUDOTASK.keys())
            num_pretask = sum(task[0] for task in PSEUDOTASK.values())

        self.num_pretask = num_pretask
        self.num_task = (self.num_task or 0) + num_pretask

        if self.verbose:
            if self.predefined_task is None:
                print(f"Using {self.num_task} user-defined tasks.")
            elif self.num_task == num_pretask:
                print(f"Using {num_pretask} predefined tasks from: {self.predefined_task}")
            else:
                print(f"Using {num_pretask} predefined tasks and {self.num_task - num_pretask} user-defined tasks.")

    @staticmethod
    def _get_param_names() -> List[str]:
        """Get parameter names for the estimator.

        Returns
        -------
        List[str]
            List of parameter names that can be used for model configuration.
        """
        return ["num_task", "predefined_task", "num_pretask"] + GNN_ENCODER_PARAMS.copy()
    
    def _get_model_params(self, checkpoint: Optional[Dict] = None) -> Dict[str, Any]:
        params = {
            "num_layer": self.num_layer,
            "hidden_size": self.hidden_size, 
            "num_task": self.num_task,
            "encoder_type": self.encoder_type,
            "drop_ratio": self.drop_ratio,
            "norm_layer": self.norm_layer,
            "readout": self.readout,
        }
        
        if checkpoint is not None:
            if "hyperparameters" not in checkpoint:
                raise ValueError("Checkpoint missing 'hyperparameters' key")
            hyperparameters = checkpoint["hyperparameters"]
            params = {k: hyperparameters.get(k, v) for k, v in params.items()}
            
        return params
    def _convert_to_pytorch_data(self, X, y=None):
        """Convert numpy arrays to PyTorch Geometric data format.
        """
        if self.verbose:
            iterator = tqdm(enumerate(X), desc="Converting molecules to graphs", total=len(X))
        else:
            iterator = enumerate(X)

        pyg_graph_list = []
        for idx, smiles_or_mol in iterator:
            if y is not None:
                properties = y[idx]
            else: 
                properties = None
            graph = graph_from_smiles(smiles_or_mol, properties, augmented_properties = self.predefined_task)
            g = Data()
            g.num_nodes = graph["num_nodes"]
            g.edge_index = torch.from_numpy(graph["edge_index"])

            del graph["num_nodes"]
            del graph["edge_index"]

            if graph["edge_feat"] is not None:
                g.edge_attr = torch.from_numpy(graph["edge_feat"])
                del graph["edge_feat"]

            if graph["node_feat"] is not None:
                g.x = torch.from_numpy(graph["node_feat"])
                del graph["node_feat"]

            if graph["y"] is not None:
                g.y = torch.from_numpy(graph["y"])
                del graph["y"]
   
            if graph["morgan"] is not None:
                g.morgan = torch.tensor(graph["morgan"], dtype=torch.int8).view(1, -1)
                del graph["morgan"]
            
            if graph["maccs"] is not None:
                g.maccs = torch.tensor(graph["maccs"], dtype=torch.int8).view(1, -1)
                del graph["maccs"]

            pyg_graph_list.append(g)

        return pyg_graph_list
    
    def _setup_optimizers(self) -> Tuple[torch.optim.Optimizer, Optional[Any]]:
        """Setup optimization components including optimizer and learning rate scheduler.
        """
        optimizer = torch.optim.Adam(
            self.model.parameters(), lr=self.learning_rate, weight_decay=self.weight_decay
        )
        scheduler = None
        if self.use_lr_scheduler:
            scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
                optimizer,
                mode="min",
                factor=self.scheduler_factor,
                patience=self.scheduler_patience,
                min_lr=1e-6,
                cooldown=0,
                eps=1e-8,
            )

        return optimizer, scheduler
    
    def fit(
        self,
        X_train: List[str],
        y_train: Optional[Union[List, np.ndarray]] = None,
    ) -> "SupervisedMolecularEncoder":
        """Fit the model to the training data with optional validation set.

        Parameters
        ----------
        X_train : List[str]
            Training set input molecular structures as SMILES strings
        y_train : Union[List, np.ndarray]
            Training set target values for representation learning
        Returns
        -------
        self : SupervisedMolecularEncoder
            Fitted estimator
        """
        user_defined_task = self.num_task - self.num_pretask
        if user_defined_task > 0:
            if y_train is None:
                raise ValueError("User-defined tasks require target values but y_train is None.")
            if y_train.shape[1] != user_defined_task:
                raise ValueError(f"Number of user-defined tasks ({user_defined_task}) must match the number of target values in y_train ({y_train.shape[1]}).")

        self._initialize_model(self.model_class)
        self.model.initialize_parameters()
        optimizer, scheduler = self._setup_optimizers()
        
        # Prepare datasets and loaders
        X_train, y_train = self._validate_inputs(X_train, y_train, return_rdkit_mol=True, num_task=self.num_task, num_pretask=self.num_pretask)
        train_dataset = self._convert_to_pytorch_data(X_train, y_train)
        train_loader = DataLoader(
            train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=0
        )
        self.fitting_loss = []
        if user_defined_task > 0:
            is_class_user = self._inspect_task_types(y_train, return_type="pt")
        else:
            is_class_user = torch.tensor([], dtype=torch.bool)

        if self.predefined_task is not None:
            is_class_predefined = torch.cat([torch.full((PSEUDOTASK[task][0],), PSEUDOTASK[task][1] == "classification", dtype=torch.bool) for task in self.predefined_task])
        else:
            is_class_predefined = torch.tensor([], dtype=torch.bool)

        is_class = torch.cat([is_class_user, is_class_predefined])
        for epoch in range(self.epochs):
            # Training phase
            train_losses = self._train_epoch(train_loader, optimizer, is_class, epoch)
            self.fitting_loss.append(np.mean(train_losses).item())
            if scheduler:
                scheduler.step(np.mean(train_losses).item())

        self.fitting_epoch = epoch
        self.is_fitted_ = True
        return self

    def _train_epoch(self, train_loader, optimizer, is_class, epoch):
        """Training logic for one epoch.

        Args:
            train_loader: DataLoader containing training data
            optimizer: Optimizer instance for model parameter updates

        Returns:
            list: List of loss values for each training step
        """
        self.model.train()
        losses = []

        iterator = (
            tqdm(train_loader, desc="Training", leave=False)
            if self.verbose
            else train_loader
        )

        for batch in iterator:
            batch = batch.to(self.device)
            optimizer.zero_grad()
            loss = self.model.compute_loss(batch, is_class)
            loss.backward()
            if self.grad_clip_value is not None:
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.grad_clip_value)
            optimizer.step()
            losses.append(loss.item())

            # Update progress bar if using tqdm
            if self.verbose:
                iterator.set_postfix({"Epoch": f"{epoch}", "Loss": f"{loss.item():.4f}"})

        return losses

    def encode(self, X: List[str], return_type: Literal["np", "pt"] = "pt") -> Union[np.ndarray, torch.Tensor]:
        """Encode molecules into vector representations.

        Parameters
        ----------
        X : List[str]
            List of SMILES strings
        return_type : Literal["np", "pt"], default="pt"
            Return type of the representations

        Returns
        -------
        representations : ndarray or torch.Tensor
            Molecular representations
        """
        self._check_is_fitted()

        # Convert to PyTorch Geometric format and create loader
        X, _ = self._validate_inputs(X, return_rdkit_mol=True)
        dataset = self._convert_to_pytorch_data(X)
        loader = DataLoader(dataset, batch_size=self.batch_size, shuffle=False)

        if self.model is None:
            raise RuntimeError("Model not initialized")

        # Generate encodings
        self.model = self.model.to(self.device)
        self.model.eval()
        encodings = []
        with torch.no_grad():
            for batch in tqdm(loader, disable=not self.verbose):
                batch = batch.to(self.device)
                out = self.model(batch)
                encodings.append(out["graph"].cpu())

        # Concatenate and convert to requested format
        encodings = torch.cat(encodings, dim=0)
        return encodings if return_type == "pt" else encodings.numpy()


import os
import torch
import torch.nn as nn
# from torch_scatter import scatter_add
from .utils import scatter_add

from ...nn import GNN_node, GNN_node_Virtualnode, MLP
from ...utils import init_weights

class GREA(nn.Module):
    def __init__(
        self,
        num_task,
        num_layer,
        gamma=0.4,
        hidden_size=300,
        gnn_type="gin-virtual",
        drop_ratio=0.5,
        norm_layer="batch_norm",
        augmented_feature=['maccs', 'morgan']
    ):
        super(GREA, self).__init__()
        gnn_name = gnn_type.split("-")[0]
        self.num_task = num_task
        self.gamma = gamma
        self.hidden_size = hidden_size

        if "virtual" in gnn_type:
            rationale_encoder = GNN_node_Virtualnode(
                2, hidden_size, JK="last", drop_ratio=drop_ratio, residual=True, gnn_name=gnn_name
            )
            self.graph_encoder = GNN_node_Virtualnode(
                num_layer,
                hidden_size,
                JK="last",
                drop_ratio=drop_ratio,
                residual=True,
                gnn_name=gnn_name,
                norm_layer=norm_layer,
            )
        else:
            rationale_encoder = GNN_node(
                2, hidden_size, JK="last", drop_ratio=drop_ratio, residual=True, gnn_name=gnn_name
            )
            self.graph_encoder = GNN_node(
                num_layer,
                hidden_size,
                JK="last",
                drop_ratio=drop_ratio,
                residual=True,
                gnn_name=gnn_name,
                norm_layer=norm_layer,
            )

        self.separator = Separator(
            rationale_encoder=rationale_encoder,
            gate_nn=torch.nn.Sequential(
                torch.nn.Linear(hidden_size, 2 * hidden_size),
                torch.nn.BatchNorm1d(2 * hidden_size),
                torch.nn.ReLU(),
                torch.nn.Dropout(),
                torch.nn.Linear(2 * hidden_size, 1),
            ),
        )
        
        graph_dim = hidden_size
        self.augmented_feature = augmented_feature
        if augmented_feature:
            if "morgan" in augmented_feature:
                graph_dim += 1024
            if "maccs" in augmented_feature:
                graph_dim += 167
        self.predictor = MLP(graph_dim, hidden_features=2 * hidden_size, out_features=num_task)

    
    def initialize_parameters(self, seed=None):
        """
        Randomly initialize all model parameters using the init_weights function.
        
        Args:
            seed (int, optional): Random seed for reproducibility. Defaults to None.
        """
        if seed is not None:
            torch.manual_seed(seed)
        
        # Initialize the main components
        init_weights(self.graph_encoder)
        init_weights(self.separator)
        init_weights(self.predictor)
        
        # Reset all parameters using PyTorch Geometric's reset function
        def reset_parameters(module):
            if hasattr(module, 'reset_parameters'):
                module.reset_parameters()
            elif hasattr(module, 'weight') and hasattr(module.weight, 'data'):
                init_weights(module)
        
        self.apply(reset_parameters)

    def _augment_graph_features(self, batched_data, h_r, h_rep):
        if self.augmented_feature:
            if 'morgan' in self.augmented_feature:
                morgan = batched_data.morgan.type_as(h_r)
                h_r = torch.cat((h_r, morgan), dim=1)
                morgans = morgan.repeat_interleave(batched_data.batch[-1]+1, dim=0)
                h_rep = torch.cat((h_rep, morgans), dim=1)
            if 'maccs' in self.augmented_feature:
                maccs = batched_data.maccs.type_as(h_r)
                h_r = torch.cat((h_r, maccs), dim=1)
                maccses = maccs.repeat_interleave(batched_data.batch[-1]+1, dim=0)
                h_rep = torch.cat((h_rep, maccses), dim=1)
        return h_r, h_rep
    
    def compute_loss(self, batched_data, criterion):
        h_node, _ = self.graph_encoder(batched_data)
        h_r, h_env, rationale_size, envir_size, _ = self.separator(batched_data, h_node)
        h_rep = (h_r.unsqueeze(1) + h_env.unsqueeze(0)).view(-1, self.hidden_size)
        h_r, h_rep = self._augment_graph_features(batched_data, h_r, h_rep)
        pred_rem = self.predictor(h_r)
        pred_rep = self.predictor(h_rep)
        loss = torch.abs(
            rationale_size / (rationale_size + envir_size)
            - self.gamma * torch.ones_like(rationale_size)
        ).mean()
        target = batched_data.y.to(torch.float32)
        is_labeled = batched_data.y == batched_data.y
        loss += criterion(pred_rem.to(torch.float32)[is_labeled], target[is_labeled]).mean()
        target_rep = batched_data.y.to(torch.float32).repeat_interleave(
            batched_data.batch[-1] + 1, dim=0
        )
        is_labeled_rep = target_rep == target_rep
        loss += criterion(pred_rep.to(torch.float32)[is_labeled_rep], target_rep[is_labeled_rep]).mean()
        return loss

    def forward(self, batched_data):
        h_node, _ = self.graph_encoder(batched_data)
        h_r, h_env, _, _, node_score = self.separator(batched_data, h_node)
        h_rep = (h_r.unsqueeze(1) + h_env.unsqueeze(0)).view(-1, self.hidden_size)
        h_r, h_rep = self._augment_graph_features(batched_data, h_r, h_rep)
        prediction = self.predictor(h_r)
        pred_rep = self.predictor(h_rep).view(h_r.size(0), -1)
        if pred_rep.size(1) > 1:
            variance = pred_rep.var(dim=-1, keepdim=True)
        else:
            variance = torch.zeros_like(pred_rep)
        num_graphs = batched_data.batch.max().item() + 1
        score_by_graph = [node_score[batched_data.batch == i].view(-1).tolist() for i in range(num_graphs)]
        return {"prediction": prediction, "variance": variance, "score": score_by_graph, "representation": h_r}

    @staticmethod
    def _disable_batchnorm_tracking(model):
        def fn(module):
            if isinstance(module, torch.nn.modules.batchnorm._BatchNorm):
                module.track_running_stats = False

        model.apply(fn)
    
    @staticmethod
    def _enable_batchnorm_tracking(model):
        def fn(module):
            if isinstance(module, torch.nn.modules.batchnorm._BatchNorm):
                module.track_running_stats = True

        model.apply(fn)

class Separator(torch.nn.Module):
    def __init__(self, rationale_encoder, gate_nn, nn=None):
        super(Separator, self).__init__()
        self.rationale_encoder = rationale_encoder
        self.gate_nn = gate_nn
        self.nn = nn

    def forward(self, batched_data, h_node, size=None):
        x, _ = self.rationale_encoder(batched_data)
        batch = batched_data.batch
        x = x.unsqueeze(-1) if x.dim() == 1 else x
        size = batch[-1].item() + 1 if size is None else size

        gate = self.gate_nn(x).view(-1, 1)
        h_node = self.nn(h_node) if self.nn is not None else h_node
        assert gate.dim() == h_node.dim() and gate.size(0) == h_node.size(0)
        gate = torch.sigmoid(gate)

        h_out = scatter_add(gate * h_node, batch, dim=0, dim_size=size)
        c_out = scatter_add((1 - gate) * h_node, batch, dim=0, dim_size=size)

        rationale_size = scatter_add(gate, batch, dim=0, dim_size=size)
        envir_size = scatter_add((1 - gate), batch, dim=0, dim_size=size)

        return h_out, c_out, rationale_size + 1e-8, envir_size + 1e-8, gate



import numpy as np
from sklearn.ensemble import RandomForestRegressor
from rdkit import Chem
from typing import List, Any
from ...utils.graph.features import getmorganfingerprint


class Oracle:
    """The default Oracle class for scoring molecules in GraphGA.
    
    This class wraps predictive models (like RandomForestRegressor) to score molecules
    based on their properties. It handles conversion of SMILES to fingerprints.
    
    Parameters
    ----------
    models : List[Any], optional
        List of trained models that implement a predict method.
        If None, RandomForestRegressors will be created when fit is called.
    num_task : int, default=1
        Number of properties to predict.
    """
    
    def __init__(self, models=None, num_task=1):
        self.models = models if models is not None else [RandomForestRegressor() for _ in range(num_task)]
        self.num_task = num_task if models is None else len(models)
        
    def _convert_to_fingerprint(self, molecules):
        """Convert SMILES or RDKit molecules to fingerprints."""
        if isinstance(molecules[0], str):
            return np.array([getmorganfingerprint(Chem.MolFromSmiles(mol)) for mol in molecules])
        else:
            return np.array([getmorganfingerprint(mol) for mol in molecules])
    
    def fit(self, X_train, y_train):
        """Fit the underlying models with training data.
        
        Parameters
        ----------
        X_train : List[str] or List[RDKit.Mol]
            Training molecules as SMILES strings or RDKit Mol objects.
        y_train : np.ndarray
            Training labels with shape (n_samples, num_task).
        
        Returns
        -------
        self : Oracle
            Fitted oracle.
        """
        X_train_fp = self._convert_to_fingerprint(X_train)
        
        for i in range(self.num_task):
            nan_mask = ~np.isnan(y_train[:, i])
            y_train_ = y_train[:, i][nan_mask]
            X_train_fp_ = X_train_fp[nan_mask]
            self.models[i].fit(X_train_fp_, y_train_)
            
        return self
    
    def __call__(self, molecules, target_values):
        """Score molecules based on their predicted properties.
        
        Parameters
        ----------
        molecules : List[str] or List[RDKit.Mol]
            Molecules to score as SMILES strings or RDKit Mol objects.
        target_values : np.ndarray,
            Scores will be based on distance to these targets.
            
        Returns
        -------
        List[float]
            Scores for each molecule.
        """
        fps = self._convert_to_fingerprint(molecules)
        scores_list = []
        
        for i, fp in enumerate(fps):
            if self.num_task == 1:
                score = self.models[0].predict([fp])[0]
                scores_list.append(float(score))
            else:
                mol_scores = []
                for idx in range(self.num_task):
                    pred = self.models[idx].predict([fp])[0]
                    
                    if target_values is not None and not np.isnan(target_values[0][idx]):
                        # Lower score for values closer to target
                        target = target_values[0][idx]
                        dist = abs(float(pred) - target) / (abs(target) + 1e-8)
                        mol_scores.append(dist)
                        
                score = np.nanmean(mol_scores)  # Lower is better when using distances
                scores_list.append(float(score))

        return scores_list


import warnings
import torch
import numpy as np
from ..utils import (
    roc_auc_score,
    accuracy_score,
    mean_absolute_error,
    mean_squared_error,
    root_mean_squared_error,
    r2_score,
)
from dataclasses import dataclass, field
from abc import ABC, abstractmethod
from typing import Optional, ClassVar, Union, List, Dict, Any, Tuple, Callable, Type
from ..base.base import BaseModel

@dataclass
class BaseMolecularPredictor(BaseModel, ABC):
    """Base class for molecular discovery estimators."""
    
    num_task: int
    task_type: str
    model_name: str = field(default="BaseMolecularPredictor")
    
    DEFAULT_METRICS: ClassVar[Dict] = {
        "classification": {"default": ("roc_auc", roc_auc_score, True)},
        "regression": {"default": ("mae", mean_absolute_error, False)},
    }

    def __post_init__(self):
        super().__post_init__()
        if self.task_type not in ["classification", "regression"]:
            raise ValueError(f"Invalid task_type: {self.task_type}")
        if self.num_task <= 0:
            raise ValueError(f"num_task must be positive, got {self.num_task}")

    def _get_param_names(self) -> List[str]:
        return super()._get_param_names() + ["num_task", "task_type"]

    @abstractmethod
    def fit(self, X_train, y_train, X_val=None, y_val=None, X_unlbl=None) -> "BaseMolecularPredictor":
        pass

    @abstractmethod
    def autofit(self, X_train, y_train, X_val=None, y_val=None, search_parameters: Optional[dict] = None, n_trials: int = 10) -> "BaseMolecularPredictor": 
        pass

    @abstractmethod
    def predict(self, X: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def _setup_optimizers(self) -> Tuple[torch.optim.Optimizer, Optional[Any]]:
        pass

    @abstractmethod
    def _train_epoch(self, train_loader, optimizer):
        pass

    @abstractmethod
    def _evaluation_epoch(self, evaluate_loader):
        pass
        
    def _setup_evaluation(
        self,
        evaluate_criterion: Optional[Union[str, Callable]],
        evaluate_higher_better: Optional[bool],
    ) -> None:
        if evaluate_criterion is None:
            default_metric = self.DEFAULT_METRICS[self.task_type]["default"]
            self.evaluate_name = default_metric[0]
            self.evaluate_criterion = default_metric[1]
            self.evaluate_higher_better = default_metric[2]
        else:
            if isinstance(evaluate_criterion, str):
                metric_map = {
                    "accuracy": (accuracy_score, True),
                    "roc_auc": (roc_auc_score, True),
                    "rmse": (root_mean_squared_error, False),
                    "mse": (mean_squared_error, False),
                    "mae": (mean_absolute_error, False),
                    "r2": (r2_score, True),
                }
                if evaluate_criterion not in metric_map:
                    raise ValueError(
                        f"Unknown metric: {evaluate_criterion}. "
                        f"Available metrics: {list(metric_map.keys())}"
                    )
                self.evaluate_name = evaluate_criterion
                self.evaluate_criterion = metric_map[evaluate_criterion][0]
                self.evaluate_higher_better = (
                    metric_map[evaluate_criterion][1]
                    if evaluate_higher_better is None
                    else evaluate_higher_better
                )
            else:
                if evaluate_higher_better is None:
                    raise ValueError(
                        "evaluate_higher_better must be specified for a custom function."
                    )
                self.evaluate_name = "custom"
                self.evaluate_criterion = evaluate_criterion
                self.evaluate_higher_better = evaluate_higher_better
    
    def _load_default_criterion(self):
        if self.task_type == "regression":
            return torch.nn.L1Loss()
        elif self.task_type == "classification":
            return torch.nn.BCEWithLogitsLoss()
        else:
            warnings.warn(
                "Unknown task type. Using L1 Loss as default. "
                "Please specify 'regression' or 'classification' for better results."
            )
            return torch.nn.L1Loss()

# import os
# import json
# import datetime
# import tempfile
# import warnings
# import torch
# import numpy as np
# from rdkit import Chem
# from ..utils import (
#     roc_auc_score,
#     accuracy_score,
#     mean_absolute_error,
#     mean_squared_error,
#     root_mean_squared_error,
#     r2_score,
# )
# from ..utils.checkpoint import LocalCheckpointManager, HuggingFaceCheckpointManager
# from ..utils.checker import MolecularInputChecker
# from dataclasses import dataclass, field
# from abc import ABC, abstractmethod
# from typing import Optional, ClassVar, Union, List, Dict, Any, Tuple, Callable, Type

# @dataclass
# class BaseMolecularPredictor(ABC):
#     """Base class for molecular discovery estimators.
    
#     This class provides a structure for implementing custom molecular discovery
#     models. It includes basic functionality for parameter management, fitting,
#     prediction, and validation.
#     """

#     num_task: int
#     task_type: str
#     model_name: str = field(default="BaseMolecularPredictor")
#     device: Optional[torch.device] = field(default=None)
#     model: Optional[torch.nn.Module] = field(default=None, init=False)

#     DEFAULT_METRICS: ClassVar[Dict] = {
#         "classification": {
#             "default": ("roc_auc", roc_auc_score, True),
#         },
#         "regression": {"default": ("mae", mean_absolute_error, False)},
#     }

#     is_fitted_: bool = field(default=False, init=False)
#     model_class: Optional[Type[torch.nn.Module]] = field(default=None, init=False)

#     def __post_init__(self):
#         if self.task_type not in ["classification", "regression"]:
#             raise ValueError(
#                 f"task_type must be one of ['classification', 'regression'], got {self.task_type}"
#             )

#         if not isinstance(self.num_task, (int, np.integer)) or self.num_task <= 0:
#             raise ValueError(f"num_task must be a positive integer, got {self.num_task}")

#         if self.device is None:
#             self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
#         elif isinstance(self.device, str):
#             self.device = torch.device(self.device)

#     def get_params(self, deep: bool = True) -> Dict[str, Any]:
#         out = dict()
#         for key in self._get_param_names():
#             value = getattr(self, key)
#             if deep and hasattr(value, "get_params"):
#                 deep_items = value.get_params().items()
#                 out.update((key + "__" + k, val) for k, val in deep_items)
#             out[key] = value
#         return out

#     def set_params(self, **params) -> "BaseMolecularPredictor":
#         if not params:
#             return self

#         valid_params = self.get_params(deep=True)
#         for key, value in params.items():
#             if key not in valid_params:
#                 raise ValueError(f"Invalid parameter {key} for predictor {self}")
#             setattr(self, key, value)
#         return self

#     @abstractmethod
#     def fit(self, X_train, y_train, X_val=None, y_val=None, X_unlbl=None) -> "BaseMolecularPredictor":
#         pass

#     @abstractmethod
#     def autofit(
#         self,
#         X_train,
#         y_train,
#         X_val=None,
#         y_val=None,
#         search_parameters: Optional[dict] = None,
#         n_trials: int = 10,
#     ) -> "BaseMolecularPredictor":
#         pass

#     @abstractmethod
#     def predict(self, X: np.ndarray) -> np.ndarray:
#         pass

#     def _setup_optimizers(self) -> Tuple[torch.optim.Optimizer, Optional[Any]]:
#         raise NotImplementedError("Child class must implement optimizer and scheduler setup.")

#     def _setup_evaluation(
#         self,
#         evaluate_criterion: Optional[Union[str, Callable]],
#         evaluate_higher_better: Optional[bool],
#     ) -> None:
#         if evaluate_criterion is None:
#             default_metric = self.DEFAULT_METRICS[self.task_type]["default"]
#             self.evaluate_name = default_metric[0]
#             self.evaluate_criterion = default_metric[1]
#             self.evaluate_higher_better = default_metric[2]
#         else:
#             if isinstance(evaluate_criterion, str):
#                 metric_map = {
#                     "accuracy": (accuracy_score, True),
#                     "roc_auc": (roc_auc_score, True),
#                     "rmse": (root_mean_squared_error, False),
#                     "mse": (mean_squared_error, False),
#                     "mae": (mean_absolute_error, False),
#                     "r2": (r2_score, True),
#                 }
#                 if evaluate_criterion not in metric_map:
#                     raise ValueError(
#                         f"Unknown metric: {evaluate_criterion}. "
#                         f"Available metrics: {list(metric_map.keys())}"
#                     )
#                 self.evaluate_name = evaluate_criterion
#                 self.evaluate_criterion = metric_map[evaluate_criterion][0]
#                 self.evaluate_higher_better = (
#                     metric_map[evaluate_criterion][1]
#                     if evaluate_higher_better is None
#                     else evaluate_higher_better
#                 )
#             else:
#                 if evaluate_higher_better is None:
#                     raise ValueError(
#                         "evaluate_higher_better must be specified for a custom function."
#                     )
#                 self.evaluate_name = "custom"
#                 self.evaluate_criterion = evaluate_criterion
#                 self.evaluate_higher_better = evaluate_higher_better

#     def _train_epoch(self, train_loader, optimizer):
#         raise NotImplementedError("Child class must implement training logic.")

#     def _evaluation_epoch(self, evaluate_loader):
#         raise NotImplementedError("Child class must implement evaluation logic.")

#     def _check_is_fitted(self) -> None:
#         if not self.is_fitted_:
#             raise AttributeError(
#                 "This instance is not fitted yet. "
#                 "Call 'fit' before using this estimator."
#             )

#     def __repr__(self, N_CHAR_MAX=700):
#         attributes = {
#             name: value
#             for name, value in sorted(self.__dict__.items())
#             if not name.startswith("_") and not callable(value)
#         }

#         def format_value(v):
#             if isinstance(v, (float, np.float32, np.float64)):
#                 return f"{v:.3g}"
#             elif isinstance(v, (list, tuple, np.ndarray)) and len(v) > 6:
#                 return f"{str(v[:3])[:-1]}, ..., {str(v[-3:])[1:]}"
#             elif isinstance(v, str) and len(v) > 50:
#                 return f"'{v[:25]}...{v[-22:]}'"
#             elif isinstance(v, dict) and len(v) > 6:
#                 items = list(v.items())
#                 return f"{dict(items[:3])}...{dict(items[-3:])}"
#             elif isinstance(v, torch.nn.Module):
#                 return f"{v.__class__.__name__}(...)"
#             elif v is None:
#                 return "None"
#             else:
#                 return repr(v)

#         class_name = self.__class__.__name__
#         attributes_str = []

#         important_attrs = ["num_task", "task_type", "model_name", "is_fitted_"]
#         for attr in important_attrs:
#             if attr in attributes:
#                 value = attributes.pop(attr)
#                 attributes_str.append(f"{attr}={format_value(value)}")

#         sorted_attrs = sorted(attributes.items())
#         attributes_str.extend(f"{name}={format_value(value)}" for name, value in sorted_attrs)

#         content = ",\n    ".join(attributes_str)
#         repr_ = f"{class_name}(\n    {content}\n)"

#         if len(repr_) > N_CHAR_MAX:
#             lines = repr_.split("\n")
#             result = [lines[0]]
#             current_length = len(lines[0])
#             for line in lines[1:-1]:
#                 if current_length + len(line) + 5 > N_CHAR_MAX:
#                     result.append("    ...")
#                     break
#                 result.append(line)
#                 current_length += len(line) + 1
#             result.append(lines[-1])
#             repr_ = "\n".join(result)

#         return repr_

#     @abstractmethod
#     def _get_model_params(self, checkpoint: Optional[Dict] = None) -> Dict[str, Any]:
#         raise NotImplementedError("Child class must implement _get_model_params.")

#     def _initialize_model(
#         self, model_class: Type[torch.nn.Module], device, checkpoint: Optional[Dict] = None
#     ) -> None:
#         """Initialize the model with parameters or a checkpoint."""
#         try:
#             model_params = self._get_model_params(checkpoint)
#             self.model = model_class(**model_params)
#             try:
#                 self.model = self.model.to(device)
#             except Exception as e:
#                 raise RuntimeError(f"Failed to move model to device {device}: {str(e)}")

#             if checkpoint is not None:
#                 try:
#                     self.model.load_state_dict(checkpoint["model_state_dict"])
#                 except Exception as e:
#                     raise ValueError(f"Failed to load model state dictionary: {str(e)}")

#         except Exception as e:
#             raise RuntimeError(f"Model initialization failed: {str(e)}")

#     def _load_default_criterion(self):
#         if self.task_type == "regression":
#             return torch.nn.L1Loss()
#         elif self.task_type == "classification":
#             return torch.nn.BCEWithLogitsLoss()
#         else:
#             warnings.warn(
#                 "Unknown task type. Using L1 Loss as default. "
#                 "Please specify 'regression' or 'classification' for better results."
#             )
#             return torch.nn.L1Loss()

#     def _validate_inputs(
#         self, X: List[str], y: Optional[Union[List, np.ndarray]] = None, return_rdkit_mol: bool = True
#     ) -> Tuple[Union[List[str], List["Chem.Mol"]], Optional[np.ndarray]]:
#         return MolecularInputChecker.validate_inputs(X, y, self.num_task, 0, return_rdkit_mol)

#     # def _validate_smiles(self, smiles: str, idx: int) -> Tuple[bool, Optional[str], Optional[Chem.Mol]]:
#     #     return MolecularInputChecker.validate_smiles(smiles, idx)

#     @staticmethod
#     def _get_param_names() -> List[str]:
#         return ["num_task", "task_type", "is_fitted_"]

#     def save_to_local(self, path: str) -> None:
#         """Save to local disk using LocalCheckpointManager."""
#         LocalCheckpointManager.save_model_to_local(self, path)

#     def load_from_local(self, path: str) -> None:
#         """Load from local disk using LocalCheckpointManager."""
#         LocalCheckpointManager.load_model_from_local(self, path)

#     def save_to_hf(
#         self,
#         repo_id: str,
#         task_id: str = "default",
#         metadata_dict: Optional[Dict[str, Any]] = None,
#         metrics: Optional[Dict[str, float]] = None,
#         commit_message: str = "Update model",
#         token: Optional[str] = None,
#         private: bool = False,
#     ) -> None:
#         HuggingFaceCheckpointManager.push_to_huggingface(
#             model_instance=self,
#             repo_id=repo_id,
#             task_id=task_id,
#             metadata_dict=metadata_dict,
#             metrics=metrics,
#             commit_message=commit_message,
#             token=token,
#             private=private,
#         )

#     def load_from_hf(self, repo_id: str, path: str) -> None:
#         """Download and load model from Hugging Face Hub."""
#         HuggingFaceCheckpointManager.load_model_from_hf(self, repo_id, path)

#     def save(self, path: Optional[str] = None, repo_id: Optional[str] = None, **kwargs) -> None:
#         """
#         Automatic save. If `repo_id` is given, push to Hugging Face Hub.
#         Otherwise, save to local path.
#         """
#         if repo_id is not None:
#             self.save_to_hf(repo_id=repo_id, **kwargs)
#         else:
#             if path is None:
#                 raise ValueError("path must be provided if repo_id is not given.")
#             self.save_to_local(path)

#     def load(self, path: str, repo_id: Optional[str] = None) -> None:
#         """
#         Automatic load. If local `path` exists, load from there. 
#         Otherwise, try loading from Hugging Face Hub if `repo_id` is given.
#         """
#         if os.path.exists(path):
#             self.load_from_local(path)
#         else:
#             if repo_id is None:
#                 raise FileNotFoundError(
#                     f"No local file found at '{path}' and no repo_id provided."
#                 )
#             self.load_from_hf(repo_id, path)



import torch
import torch.nn as nn

class LSTM(nn.Module):
    def __init__(self, num_task, input_size, hidden_size, output_size, num_layer, dropout):
        super().__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.output_size = output_size
        self.num_layer = num_layer
        self.dropout = dropout
        if num_task == 0:
            self.input_dim = 1
        else:
            self.input_dim = num_task
        self.hidden_transform = nn.Linear(self.input_dim, num_layer * hidden_size)
        self.cell_transform = nn.Linear(self.input_dim, num_layer * hidden_size)
        self.encoder = nn.Embedding(input_size, hidden_size)
        self.decoder = nn.Linear(hidden_size, output_size)

        self.rnn = nn.LSTM(hidden_size, hidden_size, batch_first=True, num_layers=num_layer, dropout=dropout)
        self.initialize_parameters()

    def initialize_parameters(self):
        # encoder / decoder
        nn.init.xavier_uniform_(self.encoder.weight)
        nn.init.xavier_uniform_(self.decoder.weight)
        nn.init.constant_(self.decoder.bias, 0)
        # RNN
        for name, param in self.rnn.named_parameters():
            if 'weight' in name:
                nn.init.orthogonal_(param)
            elif 'bias' in name:
                nn.init.constant_(param, 0)
                # LSTM remember gate bias should be initialised to 1
                # https://github.com/pytorch/pytorch/issues/750
                r_gate = param[int(0.25 * len(param)):int(0.5 * len(param))]
                nn.init.constant_(r_gate, 1)

    def forward(self, input, hidden, cell):
        embeds = self.encoder(input)
        output, (hidden, cell) = self.rnn(embeds, (hidden, cell))
        output = self.decoder(output)
        return output, hidden, cell

    def init_hidden(self, bsz, target):
        hidden = self.hidden_transform(target)
        cell = self.cell_transform(target)
        hidden = hidden.view(self.num_layer, bsz, self.hidden_size)
        cell = cell.view(self.num_layer, bsz, self.hidden_size)
        return hidden, cell
    
    def compute_loss(self, batch_data, criterion):
        ipt, tgt, y = batch_data
        hidden, cell = self.init_hidden(ipt.size(0), y)
        output, hidden, cell = self.forward(ipt, hidden, cell)
        output = output.view(output.size(0) * output.size(1), -1)
        loss = criterion(output, tgt.view(-1))
        return loss


from rdkit import Chem
from typing import Set, Any, Optional
import torch
from .smiles_char_dict import SmilesCharDictionary

def canonicalize(smiles: str, include_stereocenters=True) -> Optional[str]:
    """
    Canonicalize the SMILES strings with RDKit.
    The algorithm is detailed under https://pubs.acs.org/doi/full/10.1021/acs.jcim.5b00543
    Args:
        smiles: SMILES string to canonicalize
        include_stereocenters: whether to keep the stereochemical information in the canonical SMILES string
    Returns:
        Canonicalized SMILES string, None if the molecule is invalid.
    """

    if smiles is None:
        return None

    if len(smiles) > 0:
        mol = Chem.MolFromSmiles(smiles)

        if mol is not None:
            return Chem.MolToSmiles(mol, isomericSmiles=include_stereocenters)
        else:
            return None
    else:
        return None

def remove_duplicates(list_with_duplicates):
    """
    Removes the duplicates and keeps the ordering of the original list.
    For duplicates, the first occurrence is kept and the later occurrences are ignored.
    Args:
        list_with_duplicates: list that possibly contains duplicates
    Returns:
        A list with no duplicates.
    """
    unique_set: Set[Any] = set()
    unique_list = []
    for element in list_with_duplicates:
        if element not in unique_set:
            unique_set.add(element)
            unique_list.append(element)

    return unique_list


def rnn_start_token_vector(batch_size, device='cpu'):
    """
    Returns a vector of start tokens for LSTM.
    This vector can be used to start sampling a batch of SMILES strings.

    Args:
        batch_size: how many SMILES will be generated at the same time in LSTM
        device: cpu | cuda

    Returns:
        a tensor (batch_size x 1) containing the start token
    """
    sd = SmilesCharDictionary()
    return torch.LongTensor(batch_size, 1).fill_(sd.begin_idx).to(device)



import os
import numpy as np
from tqdm import tqdm
from typing import Optional, Union, Dict, Any, Tuple, List, Callable, Literal
import warnings

import torch
from torch_geometric.loader import DataLoader
from torch_geometric.data import Data

from .architecture import GNN
from ...base import BaseMolecularPredictor
from ...utils import graph_from_smiles
from ...utils.search import (
    DEFAULT_GNN_SEARCH_SPACES,
    suggest_parameter,
    ParameterSpec,
    ParameterType,
    parse_list_params,
)

class GNNMolecularPredictor(BaseMolecularPredictor):
    """This predictor implements a GNN model for molecular property prediction tasks.
    """
    def __init__(
        self,
        # model parameters
        num_tasks: int = 1,
        task_type: str = "classification",
        num_layer: int = 5,
        emb_dim: int = 300,
        gnn_type: str = "gin-virtual",
        drop_ratio: float = 0.5,
        norm_layer: str = "batch_norm",
        graph_pooling: str = "sum",
        # augmented features
        augmented_feature: Optional[list[Literal["morgan", "maccs"]]] = ["morgan", "maccs"],
        # training parameters
        batch_size: int = 128,
        epochs: int = 500,
        loss_criterion: Optional[Callable] = None,
        evaluate_criterion: Optional[Union[str, Callable]] = None,
        evaluate_higher_better: Optional[bool] = None,
        learning_rate: float = 0.001,
        grad_clip_value: Optional[float] = None,
        weight_decay: float = 0.0,
        patience: int = 50,
        # scheduler
        use_lr_scheduler: bool = True,
        scheduler_factor: float = 0.5,
        scheduler_patience: int = 5,
        # others
        device: Optional[str] = None,
        verbose: bool = False,
        model_name: str = "GNNMolecularPredictor",
    ): 
        super().__init__(num_tasks=num_tasks, task_type=task_type, model_name=model_name, device=device)
        # Model hyperparameters
        self.num_layer = num_layer
        self.emb_dim = emb_dim
        self.gnn_type = gnn_type
        self.drop_ratio = drop_ratio
        self.norm_layer = norm_layer
        self.graph_pooling = graph_pooling
        # Augmented features
        if augmented_feature is not None:
            valid_augmented_feature = {"morgan", "maccs"}
            invalid_fps = set(augmented_feature) - valid_augmented_feature
            if invalid_fps:
                raise ValueError(
                    f"Invalid augmented types: {invalid_fps}. "
                    f"Valid options are: {list(valid_augmented_feature)}"
                )
        self.augmented_feature = augmented_feature
        # Training parameters
        self.batch_size = batch_size
        self.epochs = epochs
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.patience = patience
        self.use_lr_scheduler = use_lr_scheduler
        self.scheduler_factor = scheduler_factor
        self.scheduler_patience = scheduler_patience
        self.grad_clip_value = grad_clip_value
        self.loss_criterion = loss_criterion if loss_criterion is not None else self._load_default_criterion()
        self._setup_evaluation(evaluate_criterion, evaluate_higher_better)

        self.fitting_loss = []
        self.fitting_epoch = 0
        self.verbose = verbose
        self.model_class = GNN

    @staticmethod
    def _get_param_names() -> List[str]:
        """Get parameter names for the estimator.

        Returns
        -------
        List[str]
            List of parameter names that can be used for model configuration.
        """
        return [
            # Model Hyperparameters
            "num_tasks",
            "task_type",
            "num_layer",
            "emb_dim",
            "gnn_type",
            "drop_ratio",
            "norm_layer",
            "graph_pooling",
            # Augmented Features
            "augmented_feature",
            # Training Parameters
            "batch_size",
            "epochs",
            "learning_rate",
            "weight_decay",
            "patience",
            "grad_clip_value",
            "loss_criterion",
            # Evaluation Parameters
            "evaluate_name",
            "evaluate_criterion",
            "evaluate_higher_better",
            # Scheduler Parameters
            "use_lr_scheduler",
            "scheduler_factor",
            "scheduler_patience",
            # Other Parameters
            "fitting_epoch",
            "fitting_loss",
            "device",
            "verbose"
        ]
    
    def _get_model_params(self, checkpoint: Optional[Dict] = None) -> Dict[str, Any]:
        """Get model parameters either from checkpoint or current instance.
        
        Parameters
        ----------
        checkpoint : Optional[Dict]
            Checkpoint containing model hyperparameters
            
        Returns
        -------
        Dict[str, Any]
            Dictionary of model parameters
            
        Raises
        ------
        ValueError
            If checkpoint contains invalid parameters
        """
        if checkpoint is not None:
            if "hyperparameters" not in checkpoint:
                raise ValueError("Checkpoint missing 'hyperparameters' key")
                
            hyperparameters = checkpoint["hyperparameters"]
            
            # Define required parameters
            required_params = {
                "num_tasks", "num_layer", "emb_dim", "gnn_type",
                "drop_ratio", "norm_layer", "graph_pooling", 'augmented_feature'
            }
            
            # Validate parameters
            # invalid_params = set(hyperparameters.keys()) - required_params
            
            # Get parameters with fallback to instance values
            return {
                "num_tasks": hyperparameters.get("num_tasks", self.num_tasks),
                "num_layer": hyperparameters.get("num_layer", self.num_layer),
                "emb_dim": hyperparameters.get("emb_dim", self.emb_dim),
                "gnn_type": hyperparameters.get("gnn_type", self.gnn_type),
                "drop_ratio": hyperparameters.get("drop_ratio", self.drop_ratio),
                "norm_layer": hyperparameters.get("norm_layer", self.norm_layer),
                "graph_pooling": hyperparameters.get("graph_pooling", self.graph_pooling),
                "augmented_feature": hyperparameters.get("augmented_feature", self.augmented_feature)
            }
        else:
            # Use current instance parameters
            return {
                "num_tasks": self.num_tasks,
                "num_layer": self.num_layer,
                "emb_dim": self.emb_dim,
                "gnn_type": self.gnn_type,
                "drop_ratio": self.drop_ratio,
                "norm_layer": self.norm_layer,
                "graph_pooling": self.graph_pooling,
                "augmented_feature": self.augmented_feature
            }
        
    def _convert_to_pytorch_data(self, X, y=None):
        """Convert numpy arrays to PyTorch Geometric data format.
        """
        if self.verbose:
            iterator = tqdm(enumerate(X), desc="Converting molecules to graphs", total=len(X))
        else:
            iterator = enumerate(X)

        pyg_graph_list = []
        for idx, smiles in iterator:
            if y is not None:
                properties = y[idx]
            else:
                properties = None
            graph = graph_from_smiles(smiles, properties, self.augmented_feature)
            g = Data()
            g.num_nodes = graph["num_nodes"]
            g.edge_index = torch.from_numpy(graph["edge_index"])

            del graph["num_nodes"]
            del graph["edge_index"]

            if graph["edge_feat"] is not None:
                g.edge_attr = torch.from_numpy(graph["edge_feat"])
                del graph["edge_feat"]

            if graph["node_feat"] is not None:
                g.x = torch.from_numpy(graph["node_feat"])
                del graph["node_feat"]

            if graph["y"] is not None:
                g.y = torch.from_numpy(graph["y"])
                del graph["y"]
   
            if graph["morgan"] is not None:
                g.morgan = torch.tensor(graph["morgan"], dtype=torch.int8).view(1, -1)
                del graph["morgan"]
            
            if graph["maccs"] is not None:
                g.maccs = torch.tensor(graph["maccs"], dtype=torch.int8).view(1, -1)
                del graph["maccs"]

            pyg_graph_list.append(g)

        return pyg_graph_list
    
    def _setup_optimizers(self) -> Tuple[torch.optim.Optimizer, Optional[Any]]:
        """Setup optimization components including optimizer and learning rate scheduler.

        Returns
        -------
        Tuple[optim.Optimizer, Optional[Any]]
            A tuple containing:
            - The configured optimizer
            - The learning rate scheduler (if enabled, else None)

        Notes
        -----
        The scheduler returned can be either _LRScheduler or ReduceLROnPlateau,
        hence we use Any as the type hint instead of optim.lr_scheduler._LRScheduler
        """
        optimizer = torch.optim.Adam(
            self.model.parameters(), lr=self.learning_rate, weight_decay=self.weight_decay
        )
        if self.grad_clip_value is not None:
            for group in optimizer.param_groups:
                group.setdefault("max_norm", self.grad_clip_value)
                group.setdefault("norm_type", 2.0)

        scheduler = None
        if self.use_lr_scheduler:
            scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
                optimizer,
                mode="min",
                factor=self.scheduler_factor,
                patience=self.scheduler_patience,
                min_lr=1e-6,
                cooldown=0,
                eps=1e-8,
            )

        return optimizer, scheduler
    
    def _get_default_search_space(self):
        """Get the default hyperparameter search space.
        
        Returns
        -------
        Dict[str, ParameterSpec]
            Dictionary mapping parameter names to their search space specifications
        """
        return DEFAULT_GNN_SEARCH_SPACES

    def autofit(
        self,
        X_train: List[str],
        y_train: Optional[Union[List, np.ndarray]],
        X_val: Optional[List[str]] = None,
        y_val: Optional[Union[List, np.ndarray]] = None,
        X_unlbl: Optional[List[str]] = None,
        search_parameters: Optional[Dict[str, ParameterSpec]] = None,
        n_trials: int = 10,
    ) -> "GNNMolecularPredictor":
        """Automatically find the best hyperparameters using Optuna optimization."""
        import optuna
        # Default search parameters
        default_search_parameters = self._get_default_search_space()
        if search_parameters is None:
            search_parameters = default_search_parameters
        else:
            # Validate search parameter keys
            invalid_params = set(search_parameters.keys()) - set(default_search_parameters.keys())
            if invalid_params:
                raise ValueError(
                    f"Invalid search parameters: {invalid_params}. "
                    f"Valid parameters are: {list(default_search_parameters.keys())}"
                )
                
        if self.verbose:
            all_params = set(self._get_param_names())
            searched_params = set(search_parameters.keys())
            non_searched_params = all_params - searched_params
            
            print("\nParameter Search Configuration:")
            print("-" * 50)
            
            print("\n Parameters being searched:")
            for param in sorted(searched_params):
                spec = search_parameters[param]
                if spec.param_type == ParameterType.CATEGORICAL:
                    print(f"  • {param}: {spec.value_range}")
                else:
                    print(f"  • {param}: [{spec.value_range[0]}, {spec.value_range[1]}]")
                    
            print("\n Fixed parameters (not being searched):")
            for param in sorted(non_searched_params):
                value = getattr(self, param, "N/A")
                print(f"  • {param}: {value}")
            
            print("\n" + "-" * 50)

            print(f"\nStarting hyperparameter optimization using {self.evaluate_name} metric")
            print(f"Direction: {'maximize' if self.evaluate_higher_better else 'minimize'}")
            print(f"Number of trials: {n_trials}")

        # Variables to track best state
        best_score = float('-inf') if self.evaluate_higher_better else float('inf')
        best_state_dict = None
        best_trial_params = None
        best_loss = None
        best_epoch = None
        
        def objective(trial):
            nonlocal best_score, best_state_dict, best_trial_params, best_loss, best_epoch
            
            # Define hyperparameters to optimize using the parameter specifications
            params = {}
            for param_name, param_spec in search_parameters.items():
                try:
                    params[param_name] = suggest_parameter(trial, param_name, param_spec)
                except Exception as e:
                    print(f"Error suggesting parameter {param_name}: {str(e)}")
                    return float('inf')
            
            try:
                # Update model parameters and train
                if "augmented_feature" in params:
                    params['augmented_feature'] = parse_list_params(params['augmented_feature'])
                self.set_params(**params)
                self.fit(X_train, y_train, X_val, y_val, X_unlbl)
                
                # Get evaluation score
                eval_data = (X_val if X_val is not None else X_train)
                eval_labels = (y_val if y_val is not None else y_train)
                eval_results = self.predict(eval_data)
                score = float(self.evaluate_criterion(eval_labels, eval_results['prediction']))
                
                # Update best state if current score is better
                is_better = (
                    score > best_score if self.evaluate_higher_better 
                    else score < best_score
                )
                
                if is_better:
                    best_score = score
                    best_state_dict = {
                        'model': self.model.state_dict(),
                        'architecture': self._get_model_params()
                    }
                    best_trial_params = params.copy()
                    best_loss = self.fitting_loss.copy()  # Added .copy() for safety
                    best_epoch = self.fitting_epoch
                
                if self.verbose:
                    print(
                        f"Trial {trial.number}: {self.evaluate_name} = {score:.4f} "
                        f"({'better' if is_better else 'worse'} than best = {best_score:.4f})"
                    )
                    print("Current parameters:")
                    for param_name, value in params.items():
                        print(f"  {param_name}: {value}")
                
                # Return score (negated if higher is better, since Optuna minimizes)
                return -score if self.evaluate_higher_better else score
                
            except Exception as e:
                print(f"Trial {trial.number} failed with error: {str(e)}")
                return float('inf')
        
        # Create study with optional output control
        optuna.logging.set_verbosity(
            optuna.logging.INFO if self.verbose else optuna.logging.WARNING
        )
        
        # Create and run study
        study = optuna.create_study(
            direction="minimize",
            study_name=f"{self.model_name}_optimization"
        )
        
        # try:
        study.optimize(
            objective,
            n_trials=n_trials,
            catch=(Exception,),
            show_progress_bar=self.verbose
        )
        
        if best_state_dict is not None:
            self.set_params(**best_trial_params)
            # Initialize model with saved architecture parameters
            self._initialize_model(self.model_class, self.device)
            # Load the saved state dict
            self.model.load_state_dict(best_state_dict['model'])
            self.fitting_loss = best_loss
            self.fitting_epoch = best_epoch
            self.is_fitted_ = True
            
            if self.verbose:
                print(f"\nOptimization completed successfully:")
                print(f"Best {self.evaluate_name}: {best_score:.4f}")
    
                eval_data = (X_val if X_val is not None else X_train)
                eval_labels = (y_val if y_val is not None else y_train)
                eval_results = self.predict(eval_data)
                score = float(self.evaluate_criterion(eval_labels, eval_results['prediction']))
                print('post score is: ', score)

                print("\nBest parameters:")
                for param, value in best_trial_params.items():
                    param_spec = search_parameters[param]
                    print(f"  {param}: {value} (type: {param_spec.param_type.value})")
                
                # Print optimization statistics
                print("\nOptimization statistics:")
                print(f"  Number of completed trials: {len(study.trials)}")
                print(f"  Number of pruned trials: {len(study.get_trials(states=[optuna.trial.TrialState.PRUNED]))}")
                print(f"  Number of failed trials: {len(study.get_trials(states=[optuna.trial.TrialState.FAIL]))}")
            else:
                raise RuntimeError("No successful trials completed during optimization")
                
        # except KeyboardInterrupt:
        #     print("\nOptimization interrupted by user. Saving best results so far...")
        #     if best_state_dict is not None:
        #         self.set_params(**best_trial_params)
        #         # CHANGE: Use same restoration logic here
        #         self._initialize_model(self.model_class, self.device)
        #         self.model.load_state_dict(best_state_dict['model'])
        #         self.fitting_loss = best_loss
        #         self.fitting_epoch = best_epoch
        #         self.is_fitted_ = True
        
        return self
    
    def fit(
        self,
        X_train: List[str],
        y_train: Optional[Union[List, np.ndarray]],
        X_val: Optional[List[str]] = None,
        y_val: Optional[Union[List, np.ndarray]] = None,
        X_unlbl: Optional[List[str]] = None,
    ) -> "GNNMolecularPredictor":
        """Fit the model to the training data with optional validation set.

        Parameters
        ----------
        X_train : List[str]
            Training set input molecular structures as SMILES strings
        y_train : Union[List, np.ndarray]
            Training set target values for property prediction
        X_val : List[str], optional
            Validation set input molecular structures as SMILES strings.
            If None, training data will be used for validation
        y_val : Union[List, np.ndarray], optional
            Validation set target values. Required if X_val is provided
        X_unlbl : List[str], optional
            Unlabeled set input molecular structures as SMILES strings.
            
        Returns
        -------
        self : GNNMolecularPredictor
            Fitted estimator
        """
        if (X_val is None) != (y_val is None):
            raise ValueError(
                "Both X_val and y_val must be provided for validation. "
                f"Got X_val={X_val is not None}, y_val={y_val is not None}"
            )

        self._initialize_model(self.model_class, self.device)
        self.model.initialize_parameters()
        self.model = self.model.to(self.device)
        optimizer, scheduler = self._setup_optimizers()
        
        # Prepare datasets and loaders
        X_train, y_train = self._validate_inputs(X_train, y_train)
        train_dataset = self._convert_to_pytorch_data(X_train, y_train)
        train_loader = DataLoader(
            train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=0
        )

        if X_val is None or y_val is None:
            val_loader = train_loader
            warnings.warn(
                "No validation set provided. Using training set for validation. "
                "This may lead to overfitting.",
                UserWarning
            )
        else:
            X_val, y_val = self._validate_inputs(X_val, y_val)
            val_dataset = self._convert_to_pytorch_data(X_val, y_val)
            val_loader = DataLoader(
                val_dataset,
                batch_size=self.batch_size,
                shuffle=False,
                num_workers=0
            )

        # Initialize training state
        self.fitting_loss = []
        self.fitting_epoch = 0
        best_state_dict = None
        best_eval = float('-inf') if self.evaluate_higher_better else float('inf')
        cnt_wait = 0

        for epoch in range(self.epochs):
            # Training phase
            train_losses = self._train_epoch(train_loader, optimizer)
            self.fitting_loss.append(np.mean(train_losses))

            # Validation phase
            current_eval = self._evaluation_epoch(val_loader)
            
            if scheduler:
                scheduler.step(current_eval)
            
            # Model selection (check if current evaluation is better)
            is_better = (
                current_eval > best_eval if self.evaluate_higher_better
                else current_eval < best_eval
            )
            
            if is_better:
                self.fitting_epoch = epoch
                best_eval = current_eval
                best_state_dict = self.model.state_dict()
                cnt_wait = 0
            else:
                cnt_wait += 1
                if cnt_wait > self.patience:
                    if self.verbose:
                        print(f"Early stopping triggered after {epoch} epochs")
                    break
            
            if self.verbose and epoch % 10 == 0:
                print(
                    f"Epoch {epoch}: Loss = {np.mean(train_losses):.4f}, "
                    f"{self.evaluate_name} = {current_eval:.4f}, "
                    f"Best {self.evaluate_name} = {best_eval:.4f}"
                )

        # Restore best model
        if best_state_dict is not None:
            self.model.load_state_dict(best_state_dict)
        else:
            warnings.warn(
                "No improvement was achieved during training. "
                "The model may not be fitted properly.",
                UserWarning
            )

        self.is_fitted_ = True
        return self

    def predict(self, X: List[str]) -> Dict[str, np.ndarray]:
        """Make predictions using the fitted model.

        Parameters
        ----------
        X : List[str]
            List of SMILES strings to make predictions for

        Returns
        -------
        Dict[str, np.ndarray]
            Dictionary containing:
                - 'prediction': Model predictions (shape: [n_samples, n_tasks])

        """
        self._check_is_fitted()

        # Convert to PyTorch Geometric format and create loader
        X, _ = self._validate_inputs(X)
        dataset = self._convert_to_pytorch_data(X)
        loader = DataLoader(dataset, batch_size=self.batch_size, shuffle=False)

        if self.model is None:
            raise RuntimeError("Model not initialized")
        # Make predictions
        self.model = self.model.to(self.device)
        self.model.eval()
        predictions = []
        with torch.no_grad():
            for batch in tqdm(loader, disable=not self.verbose):
                batch = batch.to(self.device)
                out = self.model(batch)
                predictions.append(out["prediction"].cpu().numpy())
        return {
            "prediction": np.concatenate(predictions, axis=0),
        }

    def _evaluation_epoch(
        self,
        loader: DataLoader,
    ) -> float:
        """Evaluate the model on given data.
        
        Parameters
        ----------
        loader : DataLoader
            DataLoader containing evaluation data
        train_losses : List[float]
            Training losses from current epoch
            
        Returns
        -------
        float
            Evaluation metric value (adjusted for higher/lower better)
        """
        self.model.eval()
        y_pred_list = []
        y_true_list = []
        
        with torch.no_grad():
            for batch in loader:
                batch = batch.to(self.device)
                out = self.model(batch)
                y_pred_list.append(out["prediction"].cpu().numpy())
                y_true_list.append(batch.y.cpu().numpy())
        
        y_pred = np.concatenate(y_pred_list, axis=0)
        y_true = np.concatenate(y_true_list, axis=0)
        
        # Compute metric
        metric_value = float(self.evaluate_criterion(y_true, y_pred))
        
        # Adjust metric value based on higher/lower better
        return metric_value

    def _train_epoch(self, train_loader, optimizer):
        """Training logic for one epoch.

        Args:
            train_loader: DataLoader containing training data
            optimizer: Optimizer instance for model parameter updates

        Returns:
            list: List of loss values for each training step
        """
        self.model.train()
        losses = []

        iterator = (
            tqdm(train_loader, desc="Training", leave=False)
            if self.verbose
            else train_loader
        )

        for batch in iterator:
            batch = batch.to(self.device)
            optimizer.zero_grad()

            # Forward pass and loss computation
            loss = self.model.compute_loss(batch, self.loss_criterion)

            # Backward pass
            loss.backward()

            # Compute gradient norm if gradient clipping is enabled
            if self.grad_clip_value is not None:
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.grad_clip_value)

            optimizer.step()

            losses.append(loss.item())

            # Update progress bar if using tqdm
            if self.verbose:
                iterator.set_postfix({"loss": f"{loss.item():.4f}"})

        return losses


__version__ = "0.1.0"

"""
predictor module
"""
from .predictor.grea import GREAMolecularPredictor
from .predictor.gnn import GNNMolecularPredictor
from .predictor.sgir import SGIRMolecularPredictor
from .predictor.irm import IRMMolecularPredictor
from .predictor.ssr import SSRMolecularPredictor
from .predictor.dir import DIRMolecularPredictor
from .predictor.rpgnn import RPGNNMolecularPredictor
from .predictor.bfgnn import BFGNNMolecularPredictor
from .predictor.grin import GRINMolecularPredictor
from .predictor.lstm import LSTMMolecularPredictor
from .predictor.smiles_transformer import SMILESTransformerMolecularPredictor
"""
encoder module
"""
from .encoder.attrmask import AttrMaskMolecularEncoder
from .encoder.moama import MoamaMolecularEncoder
from .encoder.graphmae import GraphMAEMolecularEncoder
from .encoder.supervised import SupervisedMolecularEncoder
from .encoder.contextpred import ContextPredMolecularEncoder
from .encoder.edgepred import EdgePredMolecularEncoder
from .encoder.infograph import InfoGraphMolecularEncoder
from .encoder.pretrained import HFPretrainedMolecularEncoder
"""
generator module
"""
from .generator.graph_dit import GraphDITMolecularGenerator
from .generator.digress import DigressMolecularGenerator
from .generator.gdss import GDSSMolecularGenerator
from .generator.graph_ga import GraphGAMolecularGenerator
from .generator.jtvae import JTVAEMolecularGenerator
from .generator.lstm import LSTMMolecularGenerator
from .generator.molgpt import MolGPTMolecularGenerator

__all__ = [
    # 'BaseMolecularPredictor',
    # predictors
    'SGIRMolecularPredictor',
    'GREAMolecularPredictor',
    'GNNMolecularPredictor',
    'IRMMolecularPredictor',
    'SSRMolecularPredictor',
    'DIRMolecularPredictor',
    'RPGNNMolecularPredictor',
    'BFGNNMolecularPredictor',
    'GRINMolecularPredictor',
    'LSTMMolecularPredictor',
    'SMILESTransformerMolecularPredictor',
    # encoders
    'SupervisedMolecularEncoder',
    'AttrMaskMolecularEncoder',
    'MoamaMolecularEncoder',
    'GraphMAEMolecularEncoder',
    'ContextPredMolecularEncoder',
    'EdgePredMolecularEncoder',
    'InfoGraphMolecularEncoder',
    'HFPretrainedMolecularEncoder',
    # generators
    'GraphDITMolecularGenerator',
    'DigressMolecularGenerator',
    'GDSSMolecularGenerator',
    'GraphGAMolecularGenerator',
    'JTVAEMolecularGenerator',
    'MolGPTMolecularGenerator',
    'LSTMMolecularGenerator',
]


import torch
import torch.nn as nn
from torch_geometric.nn import global_add_pool, global_mean_pool, global_max_pool

from ...nn import GNN_node, GNN_node_Virtualnode, GCNConv, GINConv
from ...utils import init_weights

from .utils import get_mask_indices, get_fingerprint_loss

class_criterion = torch.nn.CrossEntropyLoss()

class GNN(nn.Module):
    def __init__(
        self,
        num_layer,
        hidden_size,
        drop_ratio=0.5,
        norm_layer="batch_norm",
        encoder_type="gin-virtual",
        readout="max",
        mask_rate=0.15,
        lw_rec=0.5
    ):
        super(GNN, self).__init__()
        gnn_name = encoder_type.split("-")[0]
        self.num_atom_type = 119
        self.hidden_size = hidden_size
        self.mask_rate = mask_rate
        self.lw_rec = lw_rec

        encoder_params = {
            "num_layer": num_layer,
            "hidden_size": hidden_size,
            "JK": "last", 
            "drop_ratio": drop_ratio,
            "residual": True,
            "gnn_name": gnn_name,
            "norm_layer": norm_layer
        }
        
        # Choose encoder type based on encoder_model
        encoder_class = GNN_node_Virtualnode if "virtual" in encoder_type else GNN_node
        self.graph_encoder = encoder_class(**encoder_params)
        pooling_funcs = {
            "sum": global_add_pool,
            "mean": global_mean_pool,
            "max": global_max_pool
        }
        self.pool = pooling_funcs.get(readout)
        if self.pool is None:
            raise ValueError(f"Invalid graph pooling type {readout}.")

        self.predictor = GNN_Decoder(hidden_size, self.num_atom_type)
    
    def initialize_parameters(self, seed=None):
        """
        Randomly initialize all model parameters using the init_weights function.
        
        Args:
            seed (int, optional): Random seed for reproducibility. Defaults to None.
        """
        if seed is not None:
            torch.manual_seed(seed)
        
        # Initialize the main components
        init_weights(self.graph_encoder)
        init_weights(self.predictor)
        
        # Reset all parameters using PyTorch Geometric's reset function
        def reset_parameters(module):
            if hasattr(module, 'reset_parameters'):
                module.reset_parameters()
            elif hasattr(module, 'weight') and hasattr(module.weight, 'data'):
                init_weights(module)
        
        self.apply(reset_parameters)

    def compute_loss(self, batched_data):
        masked_node_indices = get_mask_indices(batched_data, self.mask_rate)
        batched_data.masked_node_indices = torch.tensor(masked_node_indices)
        batched_data.y = batched_data.x[masked_node_indices][:, 0]

        # mask nodes' features
        for node_idx in masked_node_indices:
            batched_data.x[node_idx] = torch.tensor([self.num_atom_type - 1] + [0] * (batched_data.x.shape[1] - 1))
    
        # generate predictions
        h_node, _ = self.graph_encoder(batched_data)
        h_rep = self.pool(h_node, batched_data.batch)
        batched_data.x = h_node
        prediction_class = self.predictor(batched_data)[masked_node_indices]
        
        # target_class = batched_data.y.to(torch.float32)
        loss_class = class_criterion(prediction_class.to(torch.float32), batched_data.y.long())
        
        fingerprint_loss = get_fingerprint_loss(batched_data.smiles, h_rep)
                
        loss = self.lw_rec * loss_class + (1 - self.lw_rec) * fingerprint_loss

        return loss

    def forward(self, batched_data):
        h_node, _ = self.graph_encoder(batched_data)
        h_rep = self.pool(h_node, batched_data.batch)
        return {"graph": h_rep, "node": h_node}
    
class GNN_Decoder(torch.nn.Module):
    def __init__(self, hidden_size, out_dim, JK = "last", gnn_name = "gin"):
        super().__init__()
        if gnn_name == 'gin':
            self.conv = GINConv(hidden_size)
        elif gnn_name == 'gcn':
            self.conv = GCNConv(hidden_size)
        else:
            raise ValueError('Undefined GNN type called {}'.format(gnn_name))
        self.dec_token = torch.nn.Parameter(torch.zeros([1, hidden_size]))
        self.enc_to_dec = torch.nn.Linear(hidden_size, hidden_size, bias=False)    
        self.activation = torch.nn.PReLU()
        self.out = torch.nn.Linear(hidden_size, out_dim) 

    def forward(self, batched_data):
        x, edge_index, edge_attr, batch = batched_data.x, batched_data.edge_index, batched_data.edge_attr, batched_data.batch
        masked_node_indices = batched_data.masked_node_indices

        x = self.activation(x)
        x = self.enc_to_dec(x)
        x[masked_node_indices] = self.dec_token.detach().expand(len(masked_node_indices), -1)

        x = self.conv(x, edge_index, edge_attr)
        out = self.out(x)
        return out


import math

import torch
import torch.nn as nn
from torch.nn.modules.dropout import Dropout
from torch.nn.modules.linear import Linear
from torch.nn.modules.normalization import LayerNorm
from torch.nn import functional as F
from torch import Tensor

from .utils import PlaceHolder
from .diffusion import assert_correctly_masked

class GraphTransformer(nn.Module):
    def __init__(self, num_layer: int, input_dim_X: int, input_dim_E: int, input_dim_y: int, hidden_size_X: int, hidden_size_E: int, hidden_size_y: int, n_head: int,dropout: float):
        super().__init__()
        self.num_layer = num_layer
        self.out_dim_X = input_dim_X
        self.out_dim_E = input_dim_E
        self.out_dim_y = input_dim_y
        act_fn_in = nn.ReLU()
        act_fn_out = nn.ReLU()

        self.mlp_in_X = nn.Sequential(nn.Linear(input_dim_X, hidden_size_X), act_fn_in,
                                      nn.Linear(hidden_size_X, hidden_size_X), act_fn_in)

        self.mlp_in_E = nn.Sequential(nn.Linear(input_dim_E, hidden_size_E), act_fn_in,
                                      nn.Linear(hidden_size_E, hidden_size_E), act_fn_in)

        self.mlp_in_y = nn.Sequential(nn.Linear(input_dim_y, hidden_size_y), act_fn_in,
                                      nn.Linear(hidden_size_y, hidden_size_y), act_fn_in)

        self.tf_layers = nn.ModuleList([XEyTransformerLayer(dx=hidden_size_X,
                                                            de=hidden_size_E,
                                                            dy=hidden_size_y,
                                                            n_head=n_head,
                                                            dim_ffX=hidden_size_X,
                                                            dim_ffE=hidden_size_E,
                                                            dropout=dropout)
                                        for i in range(num_layer)])

        self.mlp_out_X = nn.Sequential(nn.Linear(hidden_size_X, hidden_size_X), act_fn_out,
                                       nn.Linear(hidden_size_X, self.out_dim_X))

        self.mlp_out_E = nn.Sequential(nn.Linear(hidden_size_E, hidden_size_E), act_fn_out,
                                       nn.Linear(hidden_size_E, self.out_dim_E))

        self.mlp_out_y = nn.Sequential(nn.Linear(hidden_size_y, hidden_size_y), act_fn_out,
                                       nn.Linear(hidden_size_y, self.out_dim_y))

    def initialize_parameters(self, seed=None):
        if seed is not None:
            torch.manual_seed(seed)

        # Initialize transformer layers:
        def _basic_init(module):
            if isinstance(module, nn.Linear):
                torch.nn.init.xavier_uniform_(module.weight)
                if module.bias is not None:
                    nn.init.constant_(module.bias, 0)

        self.apply(_basic_init)
            
    def compute_loss(self, noisy_data, true_X, true_E, lw_X, lw_E):
        pred = self.forward(noisy_data)
        
        # Reshape predictions and targets
        true_X = torch.reshape(true_X, (-1, true_X.size(-1)))  # (bs * n, dx)
        true_E = torch.reshape(true_E, (-1, true_E.size(-1)))  # (bs * n * n, de)
        masked_pred_X = torch.reshape(pred.X, (-1, pred.X.size(-1)))  # (bs * n, dx)
        masked_pred_E = torch.reshape(pred.E, (-1, pred.E.size(-1)))   # (bs * n * n, de)

        # Remove masked rows
        mask_X = (true_X != 0.).any(dim=-1)
        mask_E = (true_E != 0.).any(dim=-1)

        flat_true_X = true_X[mask_X, :]
        flat_pred_X = masked_pred_X[mask_X, :]
        flat_true_E = true_E[mask_E, :]
        flat_pred_E = masked_pred_E[mask_E, :]
        
        # Calculate node and edge losses using cross entropy
        loss_X = F.cross_entropy(flat_pred_X, torch.argmax(flat_true_X, dim=-1)) if true_X.numel() > 0 else 0.0
        loss_E = F.cross_entropy(flat_pred_E, torch.argmax(flat_true_E, dim=-1)) if true_E.numel() > 0 else 0.0
        loss = lw_X * loss_X + lw_E * loss_E

        return loss, loss_X, loss_E

    def forward(self, noisy_data):
        X, E, y, node_mask = noisy_data['X_t'].float(), noisy_data['E_t'].float(), noisy_data['y_t'].float(), noisy_data['node_mask']
        bs, n = X.shape[0], X.shape[1]

        diag_mask = torch.eye(n)
        diag_mask = ~diag_mask.type_as(E).bool()
        diag_mask = diag_mask.unsqueeze(0).unsqueeze(-1).expand(bs, -1, -1, -1)

        X_to_out = X[..., :self.out_dim_X]
        E_to_out = E[..., :self.out_dim_E]
        y_to_out = y[..., :self.out_dim_y]

        new_E = self.mlp_in_E(E)
        new_E = (new_E + new_E.transpose(1, 2)) / 2
        after_in = PlaceHolder(X=self.mlp_in_X(X), E=new_E, y=self.mlp_in_y(y)).mask(node_mask)
        X, E, y = after_in.X, after_in.E, after_in.y

        for layer in self.tf_layers:
            X, E, y = layer(X, E, y, node_mask)

        X = self.mlp_out_X(X)
        E = self.mlp_out_E(E)
        y = self.mlp_out_y(y)

        X = (X + X_to_out)
        E = (E + E_to_out) * diag_mask
        y = y + y_to_out

        E = 1/2 * (E + torch.transpose(E, 1, 2))

        return PlaceHolder(X=X, E=E, y=y).mask(node_mask)

class XEyTransformerLayer(nn.Module):
    """ Transformer that updates node, edge and global features
        d_x: node features
        d_e: edge features
        dz : global features
        n_head: the number of heads in the multi_head_attention
        dim_feedforward: the dimension of the feedforward network model after self-attention
        dropout: dropout probablility. 0 to disable
        layer_norm_eps: eps value in layer normalizations.
    """
    def __init__(self, dx: int, de: int, dy: int, n_head: int, dim_ffX: int = 2048,
                 dim_ffE: int = 128, dim_ffy: int = 2048, dropout: float = 0.1,
                 layer_norm_eps: float = 1e-5, device=None, dtype=None) -> None:
        kw = {'device': device, 'dtype': dtype}
        super().__init__()

        self.self_attn = NodeEdgeBlock(dx, de, dy, n_head, **kw)

        self.linX1 = Linear(dx, dim_ffX, **kw)
        self.linX2 = Linear(dim_ffX, dx, **kw)
        self.normX1 = LayerNorm(dx, eps=layer_norm_eps, **kw)
        self.normX2 = LayerNorm(dx, eps=layer_norm_eps, **kw)
        self.dropoutX1 = Dropout(dropout)
        self.dropoutX2 = Dropout(dropout)
        self.dropoutX3 = Dropout(dropout)

        self.linE1 = Linear(de, dim_ffE, **kw)
        self.linE2 = Linear(dim_ffE, de, **kw)
        self.normE1 = LayerNorm(de, eps=layer_norm_eps, **kw)
        self.normE2 = LayerNorm(de, eps=layer_norm_eps, **kw)
        self.dropoutE1 = Dropout(dropout)
        self.dropoutE2 = Dropout(dropout)
        self.dropoutE3 = Dropout(dropout)

        self.lin_y1 = Linear(dy, dim_ffy, **kw)
        self.lin_y2 = Linear(dim_ffy, dy, **kw)
        self.norm_y1 = LayerNorm(dy, eps=layer_norm_eps, **kw)
        self.norm_y2 = LayerNorm(dy, eps=layer_norm_eps, **kw)
        self.dropout_y1 = Dropout(dropout)
        self.dropout_y2 = Dropout(dropout)
        self.dropout_y3 = Dropout(dropout)

        self.activation = F.relu

    def forward(self, X: Tensor, E: Tensor, y, node_mask: Tensor):
        """ Pass the input through the encoder layer.
            X: (bs, n, d)
            E: (bs, n, n, d)
            y: (bs, dy)
            node_mask: (bs, n) Mask for the src keys per batch (optional)
            Output: newX, newE, new_y with the same shape.
        """

        newX, newE, new_y = self.self_attn(X, E, y, node_mask=node_mask)

        newX_d = self.dropoutX1(newX)
        X = self.normX1(X + newX_d)

        newE_d = self.dropoutE1(newE)
        E = self.normE1(E + newE_d)

        new_y_d = self.dropout_y1(new_y)
        y = self.norm_y1(y + new_y_d)

        ff_outputX = self.linX2(self.dropoutX2(self.activation(self.linX1(X))))
        ff_outputX = self.dropoutX3(ff_outputX)
        X = self.normX2(X + ff_outputX)

        ff_outputE = self.linE2(self.dropoutE2(self.activation(self.linE1(E))))
        ff_outputE = self.dropoutE3(ff_outputE)
        E = self.normE2(E + ff_outputE)

        ff_output_y = self.lin_y2(self.dropout_y2(self.activation(self.lin_y1(y))))
        ff_output_y = self.dropout_y3(ff_output_y)
        y = self.norm_y2(y + ff_output_y)

        return X, E, y


class NodeEdgeBlock(nn.Module):
    """ Self attention layer that also updates the representations on the edges. """
    def __init__(self, dx, de, dy, n_head, **kwargs):
        super().__init__()
        assert dx % n_head == 0, f"dx: {dx} -- nhead: {n_head}"
        self.dx = dx
        self.de = de
        self.dy = dy
        self.df = int(dx / n_head)
        self.n_head = n_head

        # Attention
        self.q = Linear(dx, dx)
        self.k = Linear(dx, dx)
        self.v = Linear(dx, dx)

        # FiLM E to X
        self.e_add = Linear(de, dx)
        self.e_mul = Linear(de, dx)

        # FiLM y to E
        self.y_e_mul = Linear(dy, dx)           # Warning: here it's dx and not de
        self.y_e_add = Linear(dy, dx)

        # FiLM y to X
        self.y_x_mul = Linear(dy, dx)
        self.y_x_add = Linear(dy, dx)

        # Process y
        self.y_y = Linear(dy, dy)
        self.x_y = Xtoy(dx, dy)
        self.e_y = Etoy(de, dy)

        # Output layers
        self.x_out = Linear(dx, dx)
        self.e_out = Linear(dx, de)
        self.y_out = nn.Sequential(nn.Linear(dy, dy), nn.ReLU(), nn.Linear(dy, dy))

    def forward(self, X, E, y, node_mask):
        """
        :param X: bs, n, d        node features
        :param E: bs, n, n, d     edge features
        :param y: bs, dz           global features
        :param node_mask: bs, n
        :return: newX, newE, new_y with the same shape.
        """
        bs, n, _ = X.shape
        x_mask = node_mask.unsqueeze(-1)        # bs, n, 1
        e_mask1 = x_mask.unsqueeze(2)           # bs, n, 1, 1
        e_mask2 = x_mask.unsqueeze(1)           # bs, 1, n, 1

        # 1. Map X to keys and queries
        Q = self.q(X) * x_mask           # (bs, n, dx)
        K = self.k(X) * x_mask           # (bs, n, dx)
        assert_correctly_masked(Q, x_mask)
        # 2. Reshape to (bs, n, n_head, df) with dx = n_head * df

        Q = Q.reshape((Q.size(0), Q.size(1), self.n_head, self.df))
        K = K.reshape((K.size(0), K.size(1), self.n_head, self.df))

        Q = Q.unsqueeze(2)                              # (bs, 1, n, n_head, df)
        K = K.unsqueeze(1)                              # (bs, n, 1, n head, df)

        # Compute unnormalized attentions. Y is (bs, n, n, n_head, df)
        Y = Q * K
        Y = Y / math.sqrt(Y.size(-1))
        assert_correctly_masked(Y, (e_mask1 * e_mask2).unsqueeze(-1))

        E1 = self.e_mul(E) * e_mask1 * e_mask2                        # bs, n, n, dx
        E1 = E1.reshape((E.size(0), E.size(1), E.size(2), self.n_head, self.df))

        E2 = self.e_add(E) * e_mask1 * e_mask2                        # bs, n, n, dx
        E2 = E2.reshape((E.size(0), E.size(1), E.size(2), self.n_head, self.df))

        # Incorporate edge features to the self attention scores.
        Y = Y * (E1 + 1) + E2                  # (bs, n, n, n_head, df)

        # Incorporate y to E
        newE = Y.flatten(start_dim=3)                      # bs, n, n, dx
        ye1 = self.y_e_add(y).unsqueeze(1).unsqueeze(1)  # bs, 1, 1, de
        ye2 = self.y_e_mul(y).unsqueeze(1).unsqueeze(1)
        newE = ye1 + (ye2 + 1) * newE

        # Output E
        newE = self.e_out(newE) * e_mask1 * e_mask2      # bs, n, n, de
        assert_correctly_masked(newE, e_mask1 * e_mask2)

        # Compute attentions. attn is still (bs, n, n, n_head, df)
        softmax_mask = e_mask2.expand(-1, n, -1, self.n_head)    # bs, 1, n, 1
        attn = masked_softmax(Y, softmax_mask, dim=2)  # bs, n, n, n_head

        V = self.v(X) * x_mask                        # bs, n, dx
        V = V.reshape((V.size(0), V.size(1), self.n_head, self.df))
        V = V.unsqueeze(1)                                     # (bs, 1, n, n_head, df)

        # Compute weighted values
        weighted_V = attn * V
        weighted_V = weighted_V.sum(dim=2)

        # Send output to input dim
        weighted_V = weighted_V.flatten(start_dim=2)            # bs, n, dx

        # Incorporate y to X
        yx1 = self.y_x_add(y).unsqueeze(1)
        yx2 = self.y_x_mul(y).unsqueeze(1)
        newX = yx1 + (yx2 + 1) * weighted_V

        # Output X
        newX = self.x_out(newX) * x_mask
        assert_correctly_masked(newX, x_mask)

        # Process y based on X axnd E
        y = self.y_y(y)
        e_y = self.e_y(E)
        x_y = self.x_y(X)
        new_y = y + x_y + e_y
        new_y = self.y_out(new_y)               # bs, dy

        return newX, newE, new_y

class Xtoy(nn.Module):
    def __init__(self, dx, dy):
        """ Map node features to global features """
        super().__init__()
        self.lin = nn.Linear(4 * dx, dy)

    def forward(self, X):
        """ X: bs, n, dx. """
        m = X.mean(dim=1)
        mi = X.min(dim=1)[0]
        ma = X.max(dim=1)[0]
        std = X.std(dim=1)
        z = torch.hstack((m, mi, ma, std))
        out = self.lin(z)
        return out


class Etoy(nn.Module):
    def __init__(self, d, dy):
        """ Map edge features to global features. """
        super().__init__()
        self.lin = nn.Linear(4 * d, dy)

    def forward(self, E):
        """ E: bs, n, n, de
            Features relative to the diagonal of E could potentially be added.
        """
        m = E.mean(dim=(1, 2))
        mi = E.min(dim=2)[0].min(dim=1)[0]
        ma = E.max(dim=2)[0].max(dim=1)[0]
        std = torch.std(E, dim=(1, 2))
        z = torch.hstack((m, mi, ma, std))
        out = self.lin(z)
        return out

def masked_softmax(x, mask, **kwargs):
    if mask.sum() == 0:
        return x
    x_masked = x.clone()
    x_masked[mask == 0] = -float("inf")
    return torch.softmax(x_masked, **kwargs)